"""
Helpers for integrating bamboo in analysis workflows (git and workflow management tools)
    and improving reproducibility
"""
import dataclasses
import datetime
import enum
import logging
import os.path
import resource
import shutil
import subprocess
import urllib.parse
from collections import defaultdict
from functools import partial
from itertools import chain
from timeit import default_timer as timer

from . import utils as utils
from .analysismodules import parseRunRange

logger = logging.getLogger(__name__)


def modAbsPath(modArg):
    # Put absolute path if module is specified by file
    mod_clName = None
    if ":" in modArg:
        modArg, mod_clName = modArg.split(":")
    if os.path.isfile(modArg):
        modArg = utils.labspath(modArg)
    if mod_clName:
        modArg = ":".join((modArg, mod_clName))
    return modArg


def _callGit(args, cwd=None, stderr=None):
    return subprocess.check_output(["git"] + args, cwd=cwd, stderr=stderr).decode().strip()


hasGit = shutil.which("git") is not None
if not hasGit:
    logger.error("No git executable found")
else:
    gitVersionStr = subprocess.check_output(["git", "--version"]).decode().strip()
    logger.debug(f"Found GIT: {gitVersionStr}")
    import packaging.version
    gitVersion = packaging.version.parse(gitVersionStr.split()[2])
    if gitVersion < packaging.version.parse("2.5"):
        logger.warning(
            "The git version is below 2.5, please consider installing "
            "a more recent version to take advantage of worktrees")


class NotGitTrackedException(Exception):
    def __init__(self, path, message):
        self.path = path
        super().__init__(message)


def splitGitWorkdir(filePath):
    """ Find the top-level directory, relative path, and .git directory for a file """
    filePath = utils.labspath(os.path.expanduser(filePath))
    fDir = os.path.dirname(utils.labspath(filePath))
    if not hasGit:
        raise NotGitTrackedException(
            filePath, f"Cannot check git status of {filePath} (no git executable found)")
    try:
        cdUp = _callGit(["rev-parse", "--show-cdup"], cwd=fDir, stderr=subprocess.DEVNULL)
        topLevel = os.path.normpath(os.path.join(fDir, cdUp))
    except subprocess.CalledProcessError as ex:
        if ex.returncode == 128:
            raise NotGitTrackedException(
                filePath, f"File {filePath} is not in a git repository") from None
        raise
    fileRelPath = os.path.relpath(filePath, topLevel)
    logger.debug(f"Repository for file {filePath}: {topLevel}, relative path: {fileRelPath}")
    return topLevel, fileRelPath


def getRemotes(workdir):
    remotes = _callGit(["remote"], cwd=workdir).split()
    remInfo = {}
    for remName in remotes:
        try:
            remInfo[remName] = {
                "url": _callGit(["remote", "get-url", remName],
                                cwd=workdir, stderr=subprocess.DEVNULL),
                "url_push": _callGit(["remote", "get-url", "--push", remName],
                                     cwd=workdir, stderr=subprocess.DEVNULL)
            }
        except subprocess.CalledProcessError:  # git below 2.7
            remInfo[remName] = {
                "url": _callGit(["config", "--get", f"remote.{remName}.url"], cwd=workdir)
            }
    return remInfo


@dataclasses.dataclass
class GitVersionInfo:
    is_dirty: bool = False
    version: str = ""
    sha1: str = ""
    git_common: str = ""
    remotes: dict = dataclasses.field(default_factory=dict)
    tag: str = ""
    tag_remotes: list = dataclasses.field(default_factory=list)
    remote_branches: list = dataclasses.field(default_factory=list)
    untracked_files: list = dataclasses.field(default_factory=list)

    def asDict(self):
        return dataclasses.asdict(self)

    class Status(enum.IntEnum):
        DIRTY = 0
        LOCALCOMMIT = 1
        COMMIT = 2
        LOCALTAG = 3
        TAG = 4

    @property
    def status(self):
        if self.is_dirty:
            return GitVersionInfo.Status.DIRTY
        elif self.version == self.tag:
            if self.tag_remotes:
                return GitVersionInfo.Status.TAG
            else:
                return GitVersionInfo.Status.LOCALTAG
        else:
            if self.remote_branches:
                return GitVersionInfo.Status.COMMIT
            else:
                return GitVersionInfo.Status.LOCALCOMMIT

    POLICY_MIN = {
        "testing": Status.DIRTY,
        "committed": Status.LOCALCOMMIT,
        "tagged": Status.LOCALTAG,
        "pushed": Status.TAG,
    }

    @staticmethod
    def forDir(workdir, withRemote=True):
        callGit = partial(_callGit, cwd=workdir)
        inst = GitVersionInfo()
        inst.sha1 = callGit(["rev-parse", "--verify", "--short", "HEAD"])
        inst.version = callGit(["describe", "--tags", "--always", "--dirty"])
        if "-dirty" in inst.version:
            inst.is_dirty = True
        logger.debug(f"Version: {inst.version}, short hash: {inst.sha1}")
        gitCommon = callGit(["rev-parse", "--git-common-dir"])
        if gitCommon == "--git-common-dir":  # git below 2.5
            gitCommon = os.path.join(_callGit(["rev-parse", "--show-toplevel"],
                                              cwd=workdir, stderr=subprocess.DEVNULL), ".git")
        if not os.path.isabs(gitCommon):
            gitCommon = os.path.normpath(os.path.join(workdir, gitCommon))
        elif os.path.realpath(workdir) != workdir:  # symlinks in workdir, use them
            pWD, pGC = utils.realcommonpath(workdir, gitCommon)
            gitCommon = os.path.join(pWD, os.path.relpath(gitCommon, pGC))
        inst.git_common = gitCommon
        if withRemote:
            inst.remotes = {
                remNm: remInfo for remNm, remInfo in getRemotes(workdir).items()
                if urllib.parse.urlparse(remInfo["url"]).scheme}  # exclude local clones
            tags = callGit(["tag", "-l"]).split()
            tag = inst.version.split("-")[0]
            if tag in tags:  # set and find on remotes
                inst.tag = tag
                for remNm in inst.remotes:
                    try:
                        lst = callGit(["ls-remote", "--tags", remNm, tag], stderr=subprocess.DEVNULL)
                        if lst:
                            remHash = lst.split()[0]
                            locHash = callGit(["rev-parse", tag])
                            if remHash != locHash:
                                logger.error(
                                    f"Tag {tag} has a different hash on remote "
                                    f"({remHash} vs {locHash})! Please fix")
                            else:
                                inst.tag_remotes.append(remNm)
                    except subprocess.CalledProcessError:
                        logger.error(f"Could not check if tag {tag} is on remote {remNm}")
                if inst.tag_remotes:
                    logger.debug(f"Tag {tag} found in remote(s): {', '.join(inst.tag_remotes)}")
            # search for the commit on remote
            inst.remote_branches = [remBr.strip() for remBr in callGit(
                ["branch", "--remotes", "--no-color", "--contains", inst.sha1]).split()]
            if inst.remote_branches:
                logger.debug(
                    f"Commit {inst.sha1} found in remote branches: {', '.join(inst.remote_branches)}")
            # check for untracked
            untracked = callGit(["ls-files", "--others", "--exclude-standard"]).split()
            if untracked:
                if len(untracked) > 10:
                    logger.warning(f"{len(untracked):d} untracked files found in {workdir}")
                else:
                    logger.warning(f"Untracked files found in {workdir}: {', '.join(untracked)}")
                inst.untracked_files = list(untracked)
        return inst

    def check(self, what="", policy="pushed"):
        if what:
            what = f" for {what}"
        if self.status == GitVersionInfo.Status.DIRTY:
            logger.error(
                f"Your working tree{what} is dirty! "
                "Please commit, tag, and push for reproducible results")
        elif self.status == GitVersionInfo.Status.LOCALCOMMIT:
            logger.warning(
                f"Please tag and/or push commit {self.sha1}{what} "
                "for results that are reproducible beyond your local repository")
        elif self.status == GitVersionInfo.Status.COMMIT:
            logger.warning(
                f"Running with commit {self.sha1}{what}. "
                "Please tag (and push) for better traceability")
        elif self.status == GitVersionInfo.Status.LOCALTAG:
            logger.warning(
                f"Running with tag {self.tag}{what}. "
                "Please push for better traceability")
        elif self.status == GitVersionInfo.Status.TAG:
            logger.info(f"Running with tag {self.tag}{what}")
        if self.status < GitVersionInfo.POLICY_MIN[policy]:
            raise RuntimeError(
                f"Git status {self.status.name}{what} failed git policy '{policy}'")

# ############################################# #
# An experiment: externalize AnalysisModule.run #
# ############################################# #


def buildVersions(mod, withRemote=True, checkPolicy=None):
    """ Make a version dictionary """
    import pkg_resources
    versions = {"bamboo_version": pkg_resources.get_distribution("bamboo-hep").version}
    modFile = mod.args.module.split(":")[0]
    try:
        import importlib
        spec = importlib.util.find_spec(modFile)
        modFile = spec.origin
    except Exception:
        pass  # actually a file then
    from .workflow import splitGitWorkdir, GitVersionInfo, NotGitTrackedException
    try:
        cfgwdir, cfgFRel = splitGitWorkdir(mod.analysisConfigName)
        cfgVers = GitVersionInfo.forDir(cfgwdir, withRemote=withRemote)
        if cfgFRel in cfgVers.untracked_files:
            cfgVers.is_dirty = True
    except NotGitTrackedException:
        logger.error(
            f"Configuration file {mod.analysisConfigName} is not in a git repository, "
            "version tracking will not work")
        cfgwdir = None
        cfgFRel = utils.labspath(os.path.expanduser(mod.analysisConfigName))  # fallback
        cfgVers = GitVersionInfo(version="unknown-dirty", is_dirty=True)
    versions["config_version"] = cfgVers.asDict()
    modVers = None
    try:
        modwdir, modFRel = splitGitWorkdir(modFile)
        if cfgwdir == modwdir:
            modVers = cfgVers
            versions["module_version"] = versions["config_version"]
            if modFRel in cfgVers.untracked_files and not versions["config_version"]["is_dirty"]:
                versions["config_version"]["is_dirty"] = True
        else:
            modVers = GitVersionInfo.forDir(modwdir, withRemote=withRemote)
            if modFRel in modVers.untracked_files:
                modVers.is_dirty = True
            versions["module_version"] = modVers.asDict()
    except NotGitTrackedException:
        modFRel = utils.labspath(os.path.expanduser(modFile))  # fallback
        try:
            import inspect
            modPkg = inspect.getmodule(mod).__package__
            modVers = GitVersionInfo(
                git_common=modPkg, version=pkg_resources.get_distribution(modPkg).version)
            logger.warning(
                f"Module {modFile} is not in a git repository, "
                f"using package {modPkg} version {modVers.version} instead")
        except ValueError:
            modVers = GitVersionInfo(version="unknown-dirty", is_dirty=True)
            logger.error(
                f"Module {modFile} is not in a git repository, "
                "version tracking will not work")
        versions["module_version"] = modVers.asDict()
    versions["bambooRun_args"] = [
        # not keeping other options that do not influence the results
        f"--module={modFRel}:{mod.__class__.__name__}",
        cfgFRel,
    ] + mod.specificArgv
    if checkPolicy is None:
        checkPolicy = mod.gitPolicy
    if checkPolicy:
        if cfgVers == modVers:
            cfgVers.check(what="config and module", policy=checkPolicy)
        else:
            cfgVers.check(what="config", policy=checkPolicy)
            if isinstance(modVers, GitVersionInfo):
                cfgVers.check(what="module", policy=checkPolicy)
    return versions


def compareVersions(stored, current):
    isSame = True
    if stored["bamboo_version"] != current["bamboo_version"]:
        logger.warning(
            f"bamboo versions differ: stored {stored['bamboo_version']}, "
            f"now running {current['bamboo_version']}")
        isSame = False
    if stored["bambooRun_args"][0] != current["bambooRun_args"][0]:
        logger.error(
            f"Different bamboo module (stored {stored['bambooRun_args'][0]}, "
            f"now running {current['bambooRun_args'][0]})! This is almost certainly wrong")
        isSame = False
    if stored["bambooRun_args"][1] != current["bambooRun_args"][1]:
        logger.error(
            f"Different analysis config (stored {stored['bambooRun_args'][1]}, "
            f"now running {current['bambooRun_args'][1]})! This is almost certainly wrong")
        isSame = False
    if (len(stored["bambooRun_args"]) != len(current["bambooRun_args"]) or not
            all(sArg == cArg for sArg, cArg in zip(
                stored["bambooRun_args"][2:], current["bambooRun_args"][2:]))):
        logger.warning(
            f"Different module-specific arguments (stored '{' '.join(stored['bambooRun_args'][2:])}', "
            f"now running '{', '.join(current['bambooRun_args'][2:])}'. This is suspicious")
        isSame = False
    if stored["module_version"].get("git_common"):
        if stored["module_version"]["git_common"] != current["module_version"].get("git_common"):
            logger.error(
                f"Module from different repository (stored {stored['module_version']['git_common']}, "
                f"now running {current['module_version'].get('git_common')}). Be very careful")
            isSame = False
    if stored["module_version"]["version"] != current["module_version"]["version"]:
        logger.error(
            f'Module from different version (stored {stored["module_version"]["version"]}, '
            f'now running {current["module_version"]["version"]}). Be careful')
        isSame = False
    if stored["config_version"].get("git_common"):
        if stored["config_version"]["git_common"] != current["config_version"].get("git_common"):
            logger.error(
                f"config from different repository (stored {stored['config_version']['git_common']}, "
                f"now running {current['config_version'].get('git_common')}). Be very careful")
            isSame = False
    if stored["config_version"]["version"] != current["config_version"]["version"]:
        logger.error(
            f'Config from different version (stored {stored["config_version"]["version"]}, '
            f'now running {current["config_version"]["version"]}). Be careful')
        isSame = False
    return isSame


class SampleTask:
    """ Information about work to be done for single sample (input and output file, parameters etc.) """
    def __init__(self, name, inputFiles=None, outputFile=None, kwargs=None, config=None, resolver=None):
        self.name = name
        self._inputFiles = inputFiles
        self.outputFile = outputFile
        self.kwargs = kwargs
        self.config = config
        self._resolver = resolver

    @property
    def inputFiles(self):
        if self._inputFiles is None:
            self._inputFiles = self._resolver(self.name, self.config)
        return self._inputFiles


def downloadCertifiedLumiFiles(tasks, workdir="."):
    """ download certified lumi files (if needed) and replace in args """
    certifLumiFiles = {tsk.kwargs["certifiedLumiFile"]
                       for tsk in tasks if "certifiedLumiFile" in tsk.kwargs}
    # download if needed
    clf_downloaded = dict()
    for clfu in certifLumiFiles:
        purl = urllib.parse.urlparse(clfu)
        if purl.scheme in ("http", "https"):
            fname = os.path.join(workdir, purl.path.split("/")[-1])
            if os.path.exists(fname):
                logger.warning(f"File {fname} exists, it will not be downloaded again from {clfu}")
            else:
                subprocess.check_call(["wget", f"--directory-prefix={workdir}", clfu],
                                      stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            clf_downloaded[clfu] = utils.labspath(fname)
    # update args
    for tsk in tasks:
        if "certifiedLumiFile" in tsk.kwargs:
            clf = tsk.kwargs["certifiedLumiFile"]
            if clf in clf_downloaded:
                tsk.kwargs["certifiedLumiFile"] = clf_downloaded[clf]

    return set(clf_downloaded.keys())


def getBatchDefaults(backend):
    if backend == "slurm":
        return {
            "sbatch_time": "0-01:00",
            "sbatch_memPerCPU": "2048",
            "stageoutFiles": ["*.root"],
            "sbatch_chdir": utils.getlcwd(),
            "sbatch_additionalOptions": ["--export=ALL"],
        }
    elif backend == "htcondor":
        return {"cmd": {
            "universe": "vanilla",
            "+MaxRuntime": f"{20*60:d}",  # 20 minutes
            "getenv": "True",
        }}


def run_interact(mod):
    """
    Interactively inspect a decorated input tree

    Available variables: ``tree`` (decorated tree), ``tup`` (raw tree),
    ``noSel`` (root selection), ``backend``, ``runAndLS`` (e.g. ``(runExpr, lumiBlockExpr)``)
    (the inputs for the lumi mask), and ``op`` (:py:mod:`bamboo.treefunctions`).
    """
    tup, smpName, smpCfg = mod.getATree()
    tree, noSel, backend, runAndLS = mod.prepareTree(tup, sample=smpName, sampleCfg=smpCfg)
    import bamboo.treefunctions as op  # noqa: F401
    import IPython
    IPython.start_ipython(argv=[], user_ns=locals())


def processOne(mod, inputFiles, outputFile, tree=None, certifiedLumiFile=None, runRange=None,
               sample=None, sampleCfg=None, inputFileLists=None):
    """
    Process inputs for one sample at a time (old processTrees), from inputs to histograms and skims

    More in detail, this will load the inputs,
    call :py:meth:`~bamboo.analysismodules.HistogramsModule.prepareTree`,
    add a lumi mask if requested,
    call :py:meth:`~bamboo.analysismodules.HistogramsModule.definePlots`,
    run over all files, and write the produced histograms to the output file.

    :param inputFiles: input file names
    :param outputFile: output file name
    :param tree: key name of the tree inside the files
    :param certifiedLumiFile: lumi mask json file name
    :param runRange: run range to consider (for efficiency of the lumi mask)
    :param sample: sample name (key in the samples block of the configuration file)
    :param sampleCfg: that sample's entry in the configuration file
    :param inputFileLists: names of files with the input files
        (optional, to avoid rewriting if this already exists)
    """
    backend, plotList = mod.makeBackendAndPlotList(
        inputFiles, tree=tree, certifiedLumiFile=certifiedLumiFile, runRange=runRange,
        sample=sample, sampleCfg=sampleCfg, inputFileLists=inputFileLists)
    if mod.args.onlyprepare:
        return

    logger.info("Start backend graph construction")
    start = timer()
    backend.buildGraph(plotList)
    end = timer()
    maxrssmb = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024
    logger.info(f"Backend graph construction done in {end - start:.2f}s, max RSS: {maxrssmb:.2f}MB")
    from .dataframebackend import FullyCompiledBackend
    if mod.args.print_frequency > 0:
        if isinstance(backend, FullyCompiledBackend):
            logger.warning("Progress printing when using the compiled backend is disabled for now")
        else:
            backend.addDependency(headers="printprogress.h")
            from .root import gbl
            mod._progressPrinter = gbl.rdfhelpers.PrintProgress.addToNode(
                backend.rootDF, mod.args.print_frequency,
                (mod.args.threads if mod.args.threads else 1))

    logger.info("Starting to fill plots (and skims)")
    start = timer()
    backend.runGraph()
    end = timer()
    maxrssmb = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024
    stats = backend.writeResults(plotList, outputFile, partial(  # reorder args for partial
        (lambda outF, inFN=None, smp=None, merge=None: merge(outF, inFN, sample=smp)),
        inFN=inputFiles, smp=sample, merge=mod.mergeCounters))
    logger.info(
        f"Plots finished in {end - start:.2f}s, max RSS: {maxrssmb:.2f}MB. "
        f"{stats['nHistos']:d} histograms, {stats['nSkims']:d} skims"
        + (f", {stats['nDDHistos']:d} histograms for associated selections"
           if stats["nDDHistos"] else "")
    )

    mod.plotList = plotList  # for postprocessing


def run_worker(mod):
    """ Run as a worker: process inputs and save everything to an output file """
    args = mod.args
    if (not args.output.endswith(".root")) or os.path.isdir(args.output):
        raise RuntimeError("Output for worker processes needs to be a ROOT file")
    if args.versions_file:
        import yaml
        with open(args.versions_file) as storedVersF:
            storedVersions = yaml.load(storedVersF, yaml.SafeLoader)
        versionsDict = buildVersions(mod, withRemote=False, checkPolicy="committed")
        if not compareVersions(storedVersions, versionsDict):
            raise RuntimeError("Code version differs from submitted")
    inputFiles = mod.inputs
    inputFileLists = None
    if args.maxFiles > 0 and args.maxFiles < len(inputFiles):
        logger.warning(f"Only processing first {args.maxFiles:d} of {len(inputFiles):d} files")
        inputFiles = inputFiles[:args.maxFiles]
    else:
        inputFileLists = args.filelists
    logger.info(
        f"Worker processing sample {args.sample} with module {args.module}. "
        f"Inputs: {inputFiles}, tree: {args.treeName}, output: {args.output}, "
        f"certifiedLumiFile={args.certifiedLumiFile}, runRange={args.runRange}")
    if args.anaConfig:
        sampleCfg = mod.analysisConfig["samples"][args.sample]
    else:
        sampleCfg = None
    processOne(
        mod, inputFiles, args.output, tree=args.treeName,
        certifiedLumiFile=args.certifiedLumiFile, runRange=args.runRange,
        sample=args.sample, sampleCfg=sampleCfg, inputFileLists=inputFileLists
    )


def run_notworker(mod):
    if len(mod.args.input) != 1:
        raise RuntimeError(
            "Main process (driver or non-distributed) needs exactly one argument "
            "(analysis description YAML file)")
    anaCfgName = mod.args.input[0]
    workdir = mod.args.output
    # write or check versions
    versionsDict = buildVersions(mod)
    versFile = os.path.join(workdir, "version.yml")
    if os.path.isfile(versFile):
        import yaml
        with open(versFile) as storedVersF:
            storedVersions = yaml.load(storedVersF, yaml.SafeLoader)
        compareVersions(storedVersions, versionsDict)
    # customise analysis config, get tasks
    analysisCfg = mod.analysisConfig
    tasks = mod.getTasks(
        analysisCfg, tree=analysisCfg.get("tree", "Events"),
        resolveFiles=(mod.sampleFilesResolver if not mod.args.onlypost else None))
    resultsdir = os.path.join(workdir, "results")
    if mod.args.onlypost:
        if os.path.exists(resultsdir):
            aProblem = False
            for tsk in tasks:
                fullOutName = os.path.join(resultsdir, tsk.outputFile)
                if not os.path.exists(fullOutName):
                    logger.error(f"Output file for {tsk.name} not found ({fullOutName})")
                    aProblem = True
            if aProblem:
                logger.error(f"Not all output files were found, cannot perform post-processing")
                return
        else:
            logger.error(
                f"Results directory {resultsdir} does not exist, cannot perform post-processing")
            return
    elif mod.args.distributed == "finalize":
        tasks_notfinalized = [tsk for tsk in tasks
                              if not os.path.exists(os.path.join(resultsdir, tsk.outputFile))]
        if not tasks_notfinalized:
            logger.info(
                f"All output files were found, so no finalization was redone. "
                "If you merged the outputs of partially-done MC samples "
                "please remove them from {resultsdir} and rerun to pick up the rest.")
        else:
            def cmdMatch(ln, smpNm):
                return f" --sample={smpNm} " in ln or ln.endswith(f" --sample={smpNm}")
            from .batch import getBackend
            batchBackend = getBackend(mod.envConfig["batch"]["backend"])
            batchDir = os.path.join(workdir, "batch")
            outputs, id_noOut = batchBackend.findOutputsForCommands(
                batchDir, {tsk.name: partial(cmdMatch, smpNm=tsk.name) for tsk in tasks_notfinalized})
            if id_noOut:
                logger.error(
                    "Missing outputs for subjobs {}, so no postprocessing will be run".format(
                        ", ".join(str(sjId) for sjId in id_noOut)))
                if hasattr(batchBackend, "getResubmitCommand"):
                    resubCommand = " ".join(batchBackend.getResubmitCommand(batchDir, id_noOut))
                    logger.info(f"Resubmit with '{resubCommand}' (and possibly additional options)")
                return
            aProblem = False
            for tsk in tasks_notfinalized:
                nExpected, tskOut = outputs[tsk.name]
                if not tskOut:
                    logger.error(f"No output files for sample {tsk.name}")
                    aProblem = True
                tskOut_by_name = defaultdict(list)
                for fn in tskOut:
                    tskOut_by_name[os.path.basename(fn)].append(fn)
                for outFileName, outFiles in tskOut_by_name.items():
                    if nExpected != len(outFiles):
                        logger.error(
                            f"Not all jobs for {tsk.name} produced an output file {outFileName} "
                            f"({len(outFiles):d}/{nExpected:d} found), cannot finalize")
                        aProblem = True
                    else:
                        haddCmd = ["hadd", "-f", os.path.join(resultsdir, outFileName)] + outFiles
                        import subprocess
                        try:
                            logger.debug(
                                "Merging outputs for sample {} with {}".format(
                                    tsk.name, " ".join(haddCmd)))
                            subprocess.check_call(haddCmd, stdout=subprocess.DEVNULL)
                        except subprocess.CalledProcessError:
                            logger.error("Failed to run {}".format(" ".join(haddCmd)))
                            aProblem = True
            if aProblem:
                logger.error(
                    "Could not finalize all tasks so no post-processing will be run "
                    "(rerun in verbose mode for the full list of directories and commands)")
                return
            else:
                logger.info("All tasks finalized")
    elif not tasks:
        logger.warning("No tasks defined, skipping to postprocess")
    else:  # need to run tasks
        downloadCertifiedLumiFiles(tasks, workdir=workdir)
        if os.path.exists(resultsdir):
            logger.warning(
                f"Output directory {resultsdir} exists, previous results may be overwritten")
        else:
            os.makedirs(resultsdir)
        if os.path.isfile(versFile):
            logger.warning(
                f"Overwriting {versFile}, but if not all results are overwritten "
                "the version will not be consistent")
            versionsDict["overwritten"] = True
        import yaml
        with open(versFile, "w") as versF:
            yaml.dump(versionsDict, versF)
        # store one "skeleton" tree (for more efficient "onlypost" later on
        aTask = tasks[0]
        aFileName = aTask.inputFiles[0]
        from .root import gbl
        aFile = gbl.TFile.Open(aFileName)
        if not aFile:
            logger.warning(f"Could not open file {aFileName}, no skeleton tree will be saved")
        else:
            treeName = analysisCfg.get("tree", "Events")
            aTree = aFile.Get(treeName)
            if not aTree:
                logger.warning(
                    f"Could not get {treeName} from file {aFileName}, no skeleton tree will be saved")
            else:
                outfName = os.path.join(
                    resultsdir, "__skeleton__{}.root".format(aTask.kwargs["sample"]))
                outf = gbl.TFile.Open(outfName, "RECREATE")
                if "/" in treeName:
                    outf.mkdir("/".join(treeName.split("/")[:-1])).cd()
                _ = aTree.CloneTree(1)  # copy header and a single event
                outf.Write()
                outf.Close()
                logger.debug(f"Skeleton tree written to {outfName}")
        # run all tasks
        if mod.args.distributed == "sequential":
            for task in tasks:
                output = os.path.join(resultsdir, task.outputFile)
                logger.info(
                    f"Sequential mode: processing sample {task.kwargs['sample']} "
                    + f"with module {mod.args.module}."
                    + " ({}, {}, {}".format(
                        task.inputFiles, output,
                        ", ".join(f"{k}={v}" for k, v in task.kwargs.items())
                    ))
                if "runRange" in task.kwargs:
                    task.kwargs["runRange"] = parseRunRange(task.kwargs["runRange"])
                processOne(mod, task.inputFiles, output, sampleCfg=task.config, **task.kwargs)
            if mod.args.onlyprepare:
                logger.info(f"Only doing preparation, so no results, and skipping postprocessing")
                return
        elif mod.args.distributed == "parallel":
            from .root import gbl
            if not hasattr(gbl.ROOT.RDF, "RunGraphs"):
                raise RuntimeError("Parallel running of multiple RDF graphs requires "
                                   "ROOT::RDF::RunGraphs (v6.24/00 or higher)")
            from .dataframebackend import DebugDataframeBackend, FullyCompiledBackend
            beTasks = []
            for task in tasks:
                if "runRange" in task.kwargs:
                    task.kwargs["runRange"] = parseRunRange(task.kwargs["runRange"])
                backend, plotList = mod.makeBackendAndPlotList(
                    task.inputFiles, sampleCfg=task.config, **task.kwargs)
                if isinstance(backend, DebugDataframeBackend):
                    raise RuntimeError("DebugDataframeBackend graphs cannot be in parallel")
                elif isinstance(backend, FullyCompiledBackend):
                    # for compiled, an option could be to return the 'parallelizable' part
                    # from buildGraph, such that compilation can be done in parallel / on batch
                    raise RuntimeError("FullyCompiledBackend cannot be used in parallel (yet)")
                if plotList:
                    beTasks.append((task, backend, plotList))
            if mod.args.onlyprepare:
                logger.info(f"Only doing preparation, so no results, and skipping postprocessing")
                return
            logger.info(f"Start backend graph construction for {len(tasks):d} tasks")
            for task, backend, plotList in beTasks:
                start = timer()
                backend.buildGraph(plotList)
                end = timer()
                maxrssmb = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024
                logger.info(f"Backend graph construction for {task.name} done "
                            f"in {end - start:.2f}s, max RSS: {maxrssmb:.2f}MB")
            logger.info("Starting to fill plots (and skims)")
            from .dataframebackend import ProductHandle, SkimTreeHandle, HistoND
            someResults = gbl.std.vector["ROOT::RDF::RResultHandle>"]()
            for _, backend, _ in beTasks:  # find at least one RResultPtr per graph
                for p in chain.from_iterable(backend.products.values()):
                    if p.product is not None and p.product is not ProductHandle.NoResult:
                        if isinstance(p, SkimTreeHandle) and p.product.result is not None:
                            someResults.push_back(p.product.result)
                            break
                        elif isinstance(p, HistoND) and not isinstance(p, gbl.TH1):
                            someResults.push_back(p.product)
                            break
            start = timer()
            gbl.ROOT.RDF.RunGraphs(someResults)
            end = timer()
            logger.info(f"Plots finished in {end - start:.2f}s, max RSS: {maxrssmb:.2f}MB.")
            for task, backend, plotList in beTasks:
                stats = backend.writeResults(
                    plotList, os.path.join(resultsdir, task.outputFile), partial(
                        (lambda outF, inFN=None, smp=None, merge=None: merge(outF, inFN, sample=smp)),
                        inFN=task.inputFiles, smp=task.name, merge=mod.mergeCounters))
                logger.info(f"{task.name}: {stats['nHistos']:d} histograms, {stats['nSkims']:d} skims"
                            + (f", {stats['nDDHistos']:d} histograms for associated selections"
                               if stats["nDDHistos"] else ""))
                mod.plotList = plotList  # for postprocessing
        elif mod.args.distributed == "driver":
            # construct the list of tasks
            from .batch import (splitInChunks, writeFileList, SplitAggregationTask,
                                HaddAction, format_runtime, getBackend)
            commArgs = ([
                "bambooRun",
                f"--module={modAbsPath(mod.args.module)}",
                "--distributed=worker",
                f"--anaConfig={utils.labspath(anaCfgName)}"] + ([
                    f"--versions-file={utils.labspath(versFile)}"]
                    if mod.gitPolicy != "testing" else [])
                + mod.specificArgv
                + (["--verbose"] if mod.args.verbose else [])
                + ([f"-t {mod.args.threads}"] if mod.args.threads else [])
            )
            beTasks = []
            for tsk in tasks:
                split = 1
                if tsk.config and "split" in tsk.config:
                    split = tsk.config["split"]
                if split >= 0:
                    # at least 1 (no splitting), at most the numer of arguments (one job per input)
                    chunks = splitInChunks(
                        tsk.inputFiles, nChunks=max(1, min(split, len(tsk.inputFiles))))
                else:  # at least 1 (one job per input), at most the number of arguments (no splitting)
                    chunks = splitInChunks(
                        tsk.inputFiles, chunkLength=max(1, min(-split, len(tsk.inputFiles))))
                cmds = []
                os.makedirs(os.path.join(workdir, "infiles"), exist_ok=True)
                for i, chunk in enumerate(chunks):
                    cfn = os.path.join(workdir, "infiles", "{}_in_{:d}.txt".format(
                        tsk.kwargs["sample"], i))
                    writeFileList(chunk, cfn)
                    cmds.append(" ".join(
                        [str(a) for a in commArgs]
                        + [f"--input={utils.labspath(cfn)}", f"--output={tsk.outputFile}"]
                        + [f"--{key}={value}" for key, value in tsk.kwargs.items()]
                    ))
                beTasks.append(SplitAggregationTask(
                    cmds, finalizeAction=HaddAction(cmds, outDir=resultsdir, options=["-f"])))
            # submit to backend
            backend = mod.envConfig["batch"]["backend"]
            batchBackend = getBackend(backend)
            defaultBatchOpts = getBatchDefaults(backend)
            # make sure we request N CPUs/job when using N threads
            if mod.args.threads:
                batchBackend.configCPUReq(defaultBatchOpts, mod.args.threads)
            clusJobs = batchBackend.jobsFromTasks(
                beTasks, workdir=os.path.join(workdir, "batch"),
                batchConfig=mod.envConfig.get(backend), configOpts=defaultBatchOpts)
            for j in clusJobs:
                j.submit()
            logger.info(
                "The status of the batch jobs will be periodically checked, "
                "and the outputs merged if necessary. "
                "If only few jobs (or the monitoring loop) fail, "
                "it may be more efficient to resubmit or rerun them manually "
                "and (if necessary) merge the outputs and produce the final results "
                "either by rerunning with --driver=finalize, or manually, "
                "using the commands that will be printed if any jobs fail, "
                "and by rerunning with the --onlypost option.")
            clusMon = batchBackend.makeTasksMonitor(
                clusJobs, beTasks, interval=int(mod.envConfig["batch"].get("update", 120)))
            collectResult = clusMon.collect()  # wait for batch jobs to finish and finalize

            if any(not tsk.failedCommands for tsk in beTasks):
                # Print time report (possibly more later)
                logger.info("Average runtime for successful tasks (to further tune job splitting):")
                for tsk in beTasks:
                    if not tsk.failedCommands:
                        totTime = sum((next(
                            clus for clus in clusJobs if cmd in clus.commandList).getRuntime(cmd)
                            for cmd in tsk.commandList), datetime.timedelta())
                        nTasks = len(tsk.commandList)
                        smpName = next(arg for arg in tsk.commandList[0].split()
                                       if arg.startswith("--sample=")).split("=")[1]
                        logger.info(
                            f" - {smpName}: {format_runtime(totTime/nTasks)} "
                            f"({nTasks:d} jobs, total: {format_runtime(totTime)})")

            if not collectResult["success"]:
                # Print missing hadd actions to be done when (if) those recovery jobs succeed
                haddCmds = list(chain.from_iterable(tsk.finalizeAction.getActions()
                                for tsk in beTasks if tsk.failedCommands and tsk.finalizeAction))
                logger.error(
                    "Finalization commands to be run are:\n{}".format(
                        "\n".join(" ".join(cmd) for cmd in haddCmds)))
                logger.error(
                    "Since there were failed jobs, I'm not doing the postprocessing step. "
                    "After performing recovery actions (see above) you may run me again "
                    "with the --distributed=finalize (to merge) "
                    "or --onlypost (if merged manually) option instead.")
                return
    try:
        mod.postProcess(tasks, config=analysisCfg, workdir=workdir, resultsdir=resultsdir)
    except Exception as ex:
        logger.exception(ex)
        logger.error(
            "Exception in postprocessing. "
            "If the worker job results (e.g. histograms) were correctly produced, "
            "you do not need to resubmit them, and may recover "
            "by running with the --onlypost option instead.")



import os.path

import pytest

testData = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def nanojetargs():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016.root"))
    tup = f.Get("Events")
    tup.GetEntry(0)
    i = 0
    while tup.nMuon < 2:
        i += 1
        tup.GetEntry(i)
    RVec_float = gbl.ROOT.VecOps.RVec["float"]
    RVec_int = gbl.ROOT.VecOps.RVec["Int_t"]
    muon_pt = RVec_float(tup.Muon_pt, tup.nMuon)
    muon_eta = RVec_float(tup.Muon_eta, tup.nMuon)
    muon_phi = RVec_float(tup.Muon_phi, tup.nMuon)
    muon_mass = RVec_float(tup.Muon_mass, tup.nMuon)
    muon_charge = RVec_int(tup.Muon_charge, tup.nMuon)
    muon_nLayers = RVec_int(tup.Muon_nTrackerLayers, tup.nMuon)
    muon_genIdx = RVec_int(tup.Muon_genPartIdx, tup.nMuon)
    gen_pt = RVec_float(tup.GenPart_pt, tup.nGenPart)
    seed = 5489
    yield ((muon_pt, muon_eta, muon_phi, muon_mass, muon_charge, muon_nLayers,
            muon_genIdx, gen_pt, seed),
           (muon_pt, muon_eta, muon_phi, muon_mass, muon_charge, muon_nLayers,
            RVec_int(), RVec_float(), seed))


@pytest.fixture(scope="module")
def roccorcalc_empty():
    from bamboo.root import gbl, loadRochesterCorrectionCalculator
    loadRochesterCorrectionCalculator()
    calc = gbl.RochesterCorrectionCalculator("")
    yield calc


@pytest.fixture(scope="module")
def roccorcalc_2016():
    from bamboo.root import gbl, loadRochesterCorrectionCalculator
    loadRochesterCorrectionCalculator()
    calc = gbl.RochesterCorrectionCalculator(os.path.join(testData, "RoccoR2016.txt"))
    yield calc


def test_rochester_empty(roccorcalc_empty):
    assert roccorcalc_empty


def test_rochester_2016(roccorcalc_empty):
    assert roccorcalc_2016


def test_rochester_nano_off_mc(roccorcalc_empty, nanojetargs):
    nanojet_mc, nanojet_data = nanojetargs
    res_mc = roccorcalc_empty.produce(*nanojet_mc)
    assert res_mc


def test_rochester_nano_off_data(roccorcalc_empty, nanojetargs):
    nanojet_mc, nanojet_data = nanojetargs
    res_data = roccorcalc_empty.produce(*nanojet_data)
    assert res_data


def test_rochester_nano_mc(roccorcalc_2016, nanojetargs):
    nanojet_mc, nanojet_data = nanojetargs
    res_mc = roccorcalc_2016.produce(*nanojet_mc)
    assert res_mc


def test_rochester_nano_data(roccorcalc_2016, nanojetargs):
    nanojet_mc, nanojet_data = nanojetargs
    res_data = roccorcalc_2016.produce(*nanojet_data)
    assert res_data



import os.path

import pytest

testData = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def decoNano():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016.root"))
    tree = f.Get("Events")
    from bamboo.treedecorators import (decorateNanoAOD, NanoAODDescription,
                                       nanoRochesterCalc, nanoJetMETCalc)
    from bamboo.dataframebackend import DataframeBackend
    tup = decorateNanoAOD(
        tree, NanoAODDescription.get(
            "v5", year="2016", isMC=True,
            systVariations=[nanoRochesterCalc, nanoJetMETCalc]))
    be, noSel = DataframeBackend.create(tup)
    from bamboo.analysisutils import configureJets, configureType1MET, configureRochesterCorrection
    configureJets(tup._Jet, "AK4PFchs", isMC=True, backend=be)
    configureType1MET(tup._MET, jec="Summer16_07Aug2017_V11_MC", isMC=True, backend=be)
    configureRochesterCorrection(tup._Muon, os.path.join(testData, "RoccoR2016.txt"),
                                 isMC=True, backend=be)
    yield tup


def test_getSimpleObjects(decoNano):
    assert decoNano.Electron[0]
    assert decoNano.Muon[0]
    assert decoNano.Photon[0]
    assert decoNano.SubJet[0]
    assert decoNano.FatJet[0]
    assert decoNano.LHEScaleWeight[0]


def test_getJet(decoNano):
    assert decoNano.Jet[0]


def test_simpleRef(decoNano):
    assert decoNano.Electron[0].photon
    assert decoNano.FatJet[0].subJet1


def test_toJetRef(decoNano):
    assert decoNano.Electron[0].jet
    assert decoNano.Muon[0].jet


def test_fromJetRef(decoNano):
    jet = decoNano.Jet[0]
    assert jet.muon1
    assert jet.muon2
    assert jet.electron1
    assert jet.electron2


def test_selectElectron(decoNano):
    from bamboo import treefunctions as op
    ele1 = op.select(decoNano.Electron, lambda ele: op.AND(ele.cutBased_Sum16 >= 3, ele.pt > 15.))
    assert ele1.op and ele1[0].pt.op


def test_selectJet(decoNano):
    from bamboo import treefunctions as op
    jets = op.select(decoNano.Jet, lambda j: op.AND(j.pt > 20., op.abs(j.eta) < 2.4, j.jetId & 2))
    assert jets.op and jets[0].pt.op


def test_selectSelectElectron(decoNano):
    from bamboo import treefunctions as op
    ele1 = op.select(decoNano.Electron, lambda ele: op.AND(ele.cutBased_Sum16 >= 3, ele.pt > 15.))
    ele2 = op.select(ele1, lambda ele: op.abs(ele.eta) < 2.5)
    assert ele1.op and ele2.op and ele2[0].pt.op


def test_rangeOps_cleanJetsSortReduceCombine(decoNano):
    from bamboo import treefunctions as op
    electrons = op.select(decoNano.Electron, lambda ele: op.AND(
        ele.cutBased_Sum16 >= 3,
        ele.pt > 15.,
        op.abs(ele.eta) < 2.4))
    muons = op.select(decoNano.Muon, lambda mu: op.AND(
        mu.pt > 10.,
        op.abs(mu.eta) < 2.4,
        mu.mediumId,
        mu.pfRelIso04_all < 0.15))
    jets = op.select(decoNano.Jet, lambda j: op.AND(
        j.pt > 20.,
        op.abs(j.eta) < 2.4,
        j.jetId & 2))
    cleanedJets = op.select(jets, lambda j: op.AND(
        op.NOT(op.rng_any(electrons, lambda ele: op.deltaR(j.p4, ele.p4) < 0.3)),
        op.NOT(op.rng_any(muons, lambda mu: op.deltaR(j.p4, mu.p4) < 0.3))))
    cleanedJetsByDeepFlav = op.sort(cleanedJets, lambda jet: jet.btagDeepFlavB)
    prodBTags = op.rng_product(cleanedJets, lambda jet: jet.btagDeepB)
    osdiele = op.combine(electrons, N=2, pred=lambda ele1, ele2: ele1.charge != ele2.charge)
    assert (electrons.op and muons.op and jets.op and cleanedJets.op
            and cleanedJets[0].pt.op and cleanedJetsByDeepFlav.op
            and cleanedJetsByDeepFlav[0].pt.op and prodBTags.op
            and osdiele.op and osdiele[0][1].pt.op)



"""
Example analysis module: make a dimuon mass plot from a NanoAOD
"""
import logging
import os.path

import bamboo.plots
from bamboo.analysismodules import (DataDrivenBackgroundHistogramsModule,
                                    NanoAODHistoModule, NanoAODModule,
                                    NanoAODSkimmerModule)

logger = logging.getLogger(__name__)


class Plot(bamboo.plots.Plot):
    def produceResults(self, bareResults, fbe, key=None):
        if any("__qcdScale" in h.GetName() for h in bareResults):
            hNom = next(h for h in bareResults if "__" not in h.GetName())
            prefix = f"{hNom.GetName()}__qcdScale"
            hVar_qcdScale = [h for h in bareResults if h.GetName().startswith(prefix)]
            if not all(hv.GetNcells() == hNom.GetNcells() for hv in hVar_qcdScale):
                logger.error("Variation histograms do not have the same binning as the nominal histogram")
            elif len(hVar_qcdScale) < 2:
                logger.error("At least two variations histograms must be provided")
            else:  # make an envelope from maximum deviations
                import numpy as np
                vars_cont = np.array([[hv.GetBinContent(i) for i in range(hv.GetNcells())]
                                      for hv in hVar_qcdScale])
                hVar_up = hNom.Clone(f"{prefix}up")
                hVar_down = hNom.Clone(f"{prefix}down")
                from itertools import count
                for i, vl, vh in zip(count(), np.amin(vars_cont, axis=0), np.amax(vars_cont, axis=0)):
                    hVar_down.SetBinContent(i, vl)
                    hVar_up.SetBinContent(i, vh)
                return bareResults + [hVar_up, hVar_down]
        return bareResults


class NanoZMuMuBase(NanoAODModule):
    """ Base module for NanoAOD Z->MuMu example """
    def addArgs(self, parser):
        super().addArgs(parser)
        parser.add_argument("--backend", type=str, default="dataframe",
                            help="Backend to use, 'dataframe' (default), 'lazy', or 'compiled'")
        parser.add_argument("--postprocessed", action="store_true",
                            help="Run on postprocessed NanoAOD")

    def prepareTree(self, tree, sample=None, sampleCfg=None):
        if self.args.postprocessed:
            return self.prepare_postprocessed(tree, sample=sample, sampleCfg=sampleCfg)
        else:
            return self.prepare_ondemand(tree, sample=sample, sampleCfg=sampleCfg)

    def prepare_ondemand(self, tree, sample=None, sampleCfg=None):
        era = sampleCfg.get("era") if sampleCfg else None
        isMC = self.isMC(sample)
        metName = "METFixEE2017" if era == "2017" else "MET"
        isNotWorker = True  # for tests - more realistic: (self.args.distributed != "worker")
        # Decorate the tree
        from bamboo.treedecorators import (NanoAODDescription, CalcCollectionsGroups,
                                           nanoRochesterCalc, nanoFatJetCalc)
        nanoJetMETCalc_both = CalcCollectionsGroups(
            Jet=("pt", "mass"), changes={metName: (f"{metName}T1", f"{metName}T1Smear")},
            **{metName: ("pt", "phi")})
        nanoJetMETCalc_data = CalcCollectionsGroups(
            Jet=("pt", "mass"), changes={metName: (f"{metName}T1",)},
            **{metName: ("pt", "phi")})
        systVars = (([nanoRochesterCalc] if era == "2016" else [])
                    + [(nanoJetMETCalc_both if isMC else nanoJetMETCalc_data), nanoFatJetCalc])
        tree, noSel, be, lumiArgs = super().prepareTree(
            tree, sample=sample, sampleCfg=sampleCfg,
            description=NanoAODDescription.get(
                "v7", year=(era if era else "2016"), isMC=isMC, systVariations=systVars),
            backend=self.args.backend)
        # per-year/era options
        puWeightsFile = None
        jecTag, smearTag = None, None
        rochesterFile = None
        if era == "2016":
            rochesterFile = os.path.join(
                os.path.dirname(os.path.dirname(__file__)), "tests", "data", "RoccoR2016.txt")
            if isMC:
                puWeightsFile = os.path.join(
                    os.path.dirname(os.path.dirname(__file__)), "tests", "data", "puweights.json")
                jecTag = "Summer16_07Aug2017_V11_MC"
                smearTag = "Summer16_25nsV1_MC"
            else:
                if "2016B" in sample or "2016C" in sample or "2016D" in sample:
                    jecTag = "Summer16_07Aug2017BCD_V11_DATA"
                elif "2016E" in sample or "2016F" in sample:
                    jecTag = "Summer16_07Aug2017EF_V11_DATA"
                elif "2016G" in sample or "2016H" in sample:
                    jecTag = "Summer16_07Aug2017GH_V11_DATA"
                else:
                    raise ValueError(f"Could not deduce data JEC tag for sample {sample}")
        elif era == "2017":
            if isMC:
                jecTag = "Fall17_17Nov2017_V32_MC"
                smearTag = "Fall17_V3_MC"
            else:
                if "2017B" in sample:
                    jecTag = "Fall17_17Nov2017B_V32_DATA"
                elif "2017C" in sample:
                    jecTag = "Fall17_17Nov2017C_V32_DATA"
                elif "2017D" in sample or "2017E" in sample:
                    jecTag = "Fall17_17Nov2017DE_V32_DATA"
                elif "2017F" in sample:
                    jecTag = "Fall17_17Nov2017F_V32_DATA"
                else:
                    raise ValueError(f"Could not deduce data JEC tag for sample {sample}")
        # always-on event weights
        if isMC:
            mcWgts = [tree.genWeight]
            if puWeightsFile:
                from bamboo.analysisutils import makePileupWeight
                mcWgts.append(makePileupWeight(
                    puWeightsFile, tree.Pileup_nTrueInt,
                    nameHint="bamboo_puWeight{}".format("".join(c for c in sample if c.isalnum()))))
                # mcWgts.append(makePileupWeight((
                #     "/afs/cern.ch/user/p/piedavid/private/bamboodev/testUL.json",
                #     "Collisions16_UltraLegacy_goldenJSON"), tree.Pileup_nTrueInt, systName="pu", sel=noSel))
            else:
                logger.warning("Running on MC without pileup reweighting")
            from bamboo import treefunctions as op
            mcWgts += [
                op.systematic(op.c_float(1.), **{f"qcdScale{i:d}": tree.LHEScaleWeight[i]
                                                 for i in (0, 1, 3, 5, 7, 8)}),
                # should only be added when present
                # op.systematic(op.c_float(1.), name="psISR", up=tree.PSWeight[2], down=tree.PSWeight[0]),
                # op.systematic(op.c_float(1.), name="psFSR", up=tree.PSWeight[3], down=tree.PSWeight[1]),
            ]
            noSel = noSel.refine("mcWeight", weight=mcWgts)

        cmJMEArgs = {
            "jec": jecTag,
            "smear": smearTag,
            "splitJER": True,
            "jesUncertaintySources": ("All" if isMC else None),
            # Alternative, for regrouped
            # "jesUncertaintySources": ("Merged" if isMC else None),
            # "regroupTag": "V2",
            "mayWriteCache": isNotWorker,
            "isMC": isMC,
            "backend": be,
        }
        # configure corrections and variations
        from bamboo.analysisutils import configureJets, configureType1MET, configureRochesterCorrection
        configureJets(tree._Jet, "AK4PFchs", **cmJMEArgs)
        if isMC:
            configureType1MET(getattr(tree, f"_{metName}T1Smear"), isT1Smear=True, **cmJMEArgs)
        configureType1MET(
            getattr(tree, f"_{metName}T1"),
            enableSystematics=((lambda v: not v.startswith("jer")) if isMC else None),
            **cmJMEArgs)
        cmJMEArgs.update({"jesUncertaintySources": (["Total"] if isMC else None), "regroupTag": None})
        configureJets(tree._FatJet, "AK8PFPuppi", mcYearForFatJets=era, **cmJMEArgs)
        if era == "2016":
            configureRochesterCorrection(tree._Muon, rochesterFile, isMC=isMC, backend=be)

        return tree, noSel, be, lumiArgs

    def prepare_postprocessed(self, tree, sample=None, sampleCfg=None):
        era = sampleCfg.get("era") if sampleCfg else None
        isMC = self.isMC(sample)
        metName = "METFixEE2017" if era == "2017" else "MET"
        # Decorate the tree
        from bamboo.treedecorators import NanoAODDescription, ReadJetMETVar, NanoReadRochesterVar
        nanoReadJetMETVar_MC = ReadJetMETVar(
            "Jet", f"{metName}_T1Smear",
            bTaggers=["csvv2", "deepcsv", "deepjet", "cmva"],
            bTagWPs=["L", "M", "T", "shape"])
        nanoReadNosmearMETVar = ReadJetMETVar(None, f"{metName}_T1")
        systVars = [nanoReadNosmearMETVar]
        if era == "2016":
            systVars.append(NanoReadRochesterVar())
        if isMC:
            systVars += [nanoReadJetMETVar_MC]
        tree, noSel, be, lumiArgs = super().prepareTree(
            tree, sample=sample, sampleCfg=sampleCfg,
            description=NanoAODDescription.get(
                "v7", year=(era if era else "2016"), isMC=isMC,
                removeGroups=f"{metName}_",
                addGroups=[f"{metName}_T1_"] + ([f"{metName}_T1Smear_"] if isMC else []),
                systVariations=systVars))

        # always-on event weights
        if isMC:
            mcWgts = [tree.genWeight]
            if era == "2016":
                mcWgts.append(tree.puWeight)
            from bamboo import treefunctions as op
            mcWgts += [
                op.systematic(
                    op.c_float(1.), name="qcdScale",
                    **{f"qcdScale{i:d}": tree.LHEScaleWeight[i] for i in (0, 1, 3, 5, 7, 8)}),
                # should only be added when present
                # op.systematic(op.c_float(1.), name="psISR", up=tree.PSWeight[2], down=tree.PSWeight[0]),
                # op.systematic(op.c_float(1.), name="psFSR", up=tree.PSWeight[3], down=tree.PSWeight[1]),
            ]
            noSel = noSel.refine("mcWeight", weight=mcWgts)

        return tree, noSel, be, lumiArgs


class NanoZMuMu(NanoZMuMuBase, DataDrivenBackgroundHistogramsModule):
    """ Example module: Z->MuMu histograms from NanoAOD """
    def definePlots(self, t, noSel, sample=None, sampleCfg=None):
        from bamboo.plots import CutFlowReport, SummedPlot, Skim
        from bamboo.plots import EquidistantBinning as EqB
        from bamboo import treefunctions as op
        from bamboo.analysisutils import forceDefine

        era = sampleCfg.get("era") if sampleCfg else None

        plots = []
        cfr = CutFlowReport("yields", recursive=True)
        plots.append(cfr)

        muons = op.select(t.Muon, lambda mu: op.AND(mu.pt > 20., op.abs(mu.eta) < 2.4))
        electrons = op.select(t.Electron, lambda el: op.AND(el.pt > 20, op.abs(el.eta) < 2.5))

        twoMuSel = noSel.refine("twoMuons", cut=[op.rng_len(muons) > 1])
        cfr.add(twoMuSel, "With two muons")
        cfr.add(twoMuSel, "With two leptons")
        plots.append(Plot.make1D(
            "dimu_M", op.invariant_mass(muons[0].p4, muons[1].p4), twoMuSel, EqB(100, 20., 120.),
            title="Dimuon invariant mass", plotopts={"show-overflow": False}))
        if self.args.backend != "compiled":
            plots.append(Skim(
                "muSkim", {
                    "nSelMuons": op.static_cast("UInt_t", op.rng_len(muons)),  # TBranch doesn't accept size_t
                    "selMuons_i": muons.idxs,
                    "selMu_miniPFRelIsoNeu": op.map(
                        muons, lambda mu: mu.miniPFRelIso_all - mu.miniPFRelIso_chg)
                }, twoMuSel,
                keepOriginal=[Skim.KeepRegex("PV_.*"), "nOtherPV", Skim.KeepRegex("OtherPV_.*")]))

        # evaluate jet and MET for all events passing twoMuSel
        # more optimization will be needed with systematics etc.
        metName = "METFixEE2017" if era == "2017" else "MET"
        if not self.args.postprocessed:
            forceDefine(t._Jet.calcProd, twoMuSel)
            forceDefine(getattr(t, f"_{metName}T1").calcProd, twoMuSel)
            if self.isMC(sample):
                forceDefine(getattr(t, f"_{metName}T1Smear").calcProd, twoMuSel)
        else:
            metName += "_"

        jets_noclean = op.select(t.Jet, lambda j: op.AND(j.jetId & 0x2, op.abs(j.eta) < 2.4, j.pt > 20.))
        jets = op.sort(
            op.select(jets_noclean, lambda j: op.AND(
                op.NOT(op.rng_any(muons, lambda l: op.deltaR(l.p4, j.p4) < 0.4)),
                op.NOT(op.rng_any(electrons, lambda l: op.deltaR(l.p4, j.p4) < 0.4))
            )), lambda j: -j.pt)

        plots.append(Plot.make1D(
            "nJets", op.rng_len(jets), twoMuSel, EqB(10, 0., 10.), title="Number of jets"))

        twoMuTwoJetSel = twoMuSel.refine("twoMuonsTwoJets", cut=[op.rng_len(jets) > 1])
        cfr.add(twoMuTwoJetSel, "With two muons and two jets")

        leadjpt = Plot.make1D(
            "leadJetPT", jets[0].pt, twoMuTwoJetSel, EqB(50, 0., 250.), title="Leading jet PT")
        # leadjphi = Plot.make1D(
        #     "leadJetPHI", op.Phi_mpi_pi(jets[0].phi), twoMuTwoJetSel, EqB(50, -3.142, 3.142),
        #     title="Leading jet PHI")
        subleadjpt = Plot.make1D(
            "subleadJetPT", jets[1].pt, twoMuTwoJetSel, EqB(50, 0., 250.), title="Subleading jet PT")
        plots += [leadjpt, subleadjpt]
        plots.append(SummedPlot(
            "twoLeadJetPT", [leadjpt, subleadjpt], xTitle="Leading two jet PTs"))
        metT1 = getattr(t, f"{metName}T1")
        if self.isMC(sample):
            metT1Smear = getattr(t, f"{metName}T1Smear")
        else:
            metT1Smear = metT1
        plots.append(Plot.make1D(
            "METT1", metT1.pt, twoMuTwoJetSel, EqB(50, 0., 250.), title="METT1 PT"))
        plots.append(Plot.make1D(
            "METT1Smear", metT1Smear.pt, twoMuTwoJetSel, EqB(50, 0., 250.), title="METT1Smear PT"))

        deepCSVFile = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                   "tests", "data", "DeepCSV_2016LegacySF_V1.csv")
        if os.path.exists(deepCSVFile):  # protection for CI tests
            from bamboo.scalefactors import BtagSF
            sf_deepcsv = BtagSF("csv", deepCSVFile, wp="Loose", sysType="central",
                                otherSysTypes=("up", "down"), measurementType="comb",
                                sel=noSel, uName=sample)
            # for reshaping: add getters={"Discri": lambda j : j.btagDeepB}

            bJets_DeepCSVLoose = op.select(jets, lambda j: j.btagDeepB > 0.2217)
            bTagSel = twoMuTwoJetSel.refine(
                "twoMuonsTwoJetsB", cut=[op.rng_len(bJets_DeepCSVLoose) > 0],
                weight=(sf_deepcsv(bJets_DeepCSVLoose[0]) if self.isMC(sample) else None))
            cfr.add(bTagSel, "With two muons a b-tag")
            plots.append(Plot.make1D(
                "bjetpt", bJets_DeepCSVLoose[0].pt, bTagSel, EqB(50, 0., 250.), title="B-jet pt"))

        from bamboo.scalefactors import get_scalefactor, binningVariables_nano
        elIDSF = get_scalefactor(
            "lepton",
            os.path.join(os.path.dirname(os.path.dirname(__file__)),
                         "tests", "data", "Electron_EGamma_SF2D_loose_moriond17.json"),
            isElectron=True,
            paramDefs=binningVariables_nano)
        if self.isMC(sample):
            genHardElectrons = op.select(t.GenPart, lambda gp: op.AND(
                (gp.statusFlags & (0x1 << 7)),
                op.abs(gp.pdgId) == 11))
            plots.append(Plot.make1D(
                "nGenElectrons", op.rng_len(genHardElectrons), noSel, EqB(5, 0., 5.),
                title="Number of electrons/positrons in the matrix element"))
            from bamboo.plots import LateSplittingSelection
            noSel = LateSplittingSelection.create(
                noSel, "splitByGenEl", keepInclusive=True, splitCuts={
                    "2El": op.rng_len(genHardElectrons) == 2,
                    "no2El": op.rng_len(genHardElectrons) != 2
                })
        twoElSel = noSel.refine(
            "twoElectrons", cut=[op.rng_len(electrons) > 1],
            weight=[elIDSF(electrons[i]) for i in range(2)])
        cfr.add(twoElSel, "With two leptons")
        plots.append(Plot.make1D(
            "Melel", op.invariant_mass(electrons[0].p4, electrons[1].p4), twoElSel,
            EqB(100, 20., 120.), title="Dielectron invariant mass",
            plotopts={"show-overflow": False}))

        return plots


class SkimNanoZMuMu(NanoZMuMuBase, NanoAODSkimmerModule):
    def defineSkimSelection(self, tree, noSel, sample=None, sampleCfg=None):
        from bamboo import treefunctions as op
        muons = op.select(tree.Muon, lambda mu: op.AND(mu.pt > 20., op.abs(mu.eta) < 2.4))
        hasTwoMu = noSel.refine("hasTwoMu", cut=(op.rng_len(muons) >= 2))
        varsToKeep = {"nMuon": None, "Muon_eta": None, "Muon_pt": None}  # from input file
        varsToKeep["nSelMuons"] = op.static_cast("UInt_t", op.rng_len(muons))  # TBranch doesn't accept size_t
        varsToKeep["selMuons_i"] = muons.idxs
        varsToKeep["selMu_miniPFRelIsoNeu"] = op.map(
            muons, lambda mu: mu.miniPFRelIso_all - mu.miniPFRelIso_chg)
        return hasTwoMu, varsToKeep


class SkimNanoAOD(NanoAODSkimmerModule):
    def prepareTree(self, tree, sample=None, sampleCfg=None):
        era = sampleCfg.get("era") if sampleCfg else None
        from bamboo.treedecorators import NanoAODDescription
        return super(NanoAODSkimmerModule, self).prepareTree(
            tree, sample=sample, sampleCfg=sampleCfg,
            description=NanoAODDescription.get("v7", year=(era if era else "2016")))


class MinimalNanoZMuMu(NanoAODHistoModule):
    def addArgs(self, parser):
        super().addArgs(parser)
        parser.add_argument(
            "--backend", type=str, default="dataframe",
            help="Backend to use, 'dataframe' (default), 'lazy', or 'compiled'")

    def prepareTree(self, tree, sample=None, sampleCfg=None):
        from bamboo.treedecorators import NanoAODDescription
        return super().prepareTree(
            tree, sample=sample, sampleCfg=sampleCfg,
            description=NanoAODDescription.get(
                "v5", year="2016", isMC=self.isMC(sample)), backend=self.args.backend)

    def definePlots(self, t, noSel, sample=None, sampleCfg=None):
        from bamboo.plots import Plot
        from bamboo.plots import EquidistantBinning as EqB
        from bamboo import treefunctions as op
        if self.isMC(sample):
            noSel = noSel.refine("mcWeight", weight=t.genWeight)
        plots = []
        muons = op.select(
            t.Muon, lambda mu: op.AND(
                mu.mediumId,
                mu.pfRelIso03_all < 0.4,
                mu.pt > 15.))
        twoMuSel = noSel.refine("has2mu", cut=(op.rng_len(muons) > 1))
        plots.append(Plot.make1D(
            "dimuM", (muons[0].p4 + muons[1].p4).M(), twoMuSel,
            EqB(100, 20., 120.), title="Invariant mass"))
        return plots



"""
Expressions are constructed by executing python code on decorated versions of
decorated trees. The :py:mod:`bamboo.treedecorators` module contains helper
methods to do so for commonly used formats, e.g. :py:func:`~.decorateNanoAOD`
for CMS NanoAOD.
"""

import logging
from collections import defaultdict
from functools import partial
from itertools import chain

from . import treefunctions as tf
from . import treeoperations as top
from . import treeproxies as tpx

logger = logging.getLogger(__name__)


def allLeafs(branch):
    # Recursively collect TTree leaves (TLeaf and TBranchElement)
    from .root import gbl
    for br in branch.GetListOfBranches():
        if isinstance(br, gbl.TBranchElement):
            yield br
        else:
            yield from br.GetListOfLeaves()


def normVarName(varName):
    # Normalize variation name: if ending in up or down, make sure this part has no capitals (for plotIt)
    if len(varName) >= 2 and varName[-2:].upper() == "UP":
        return f"{varName[:-2]}up"
    elif len(varName) >= 4 and varName[-4:].upper() == "DOWN":
        return f"{varName[:-4]}down"
    else:
        return varName

# Attribute classes (like property) to customize the proxy classes


class proxy:  # the default one
    def __init__(self, op):
        self.op = op

    def __get__(self, inst, cls):
        return self.op.result


class funProxy:  # the generic one
    def __init__(self, fun):
        self.fun = fun

    def __get__(self, inst, cls):
        return self.fun(inst)


class itemProxy:
    def __init__(self, op):
        self.op = op

    def __get__(self, inst, cls):
        return self.op[inst._idx]


class itemRefProxy:
    def __init__(self, op, getTarget):
        self.op = op
        self.getTarget = getTarget

    def __get__(self, inst, cls):
        return self.getTarget(inst)[self.op[inst._idx]]


class itemObjProxy:  # re-construct an object that was split in arrays
    def __init__(self, typeName, args):
        self.typeName = typeName
        self.args = args

    def __get__(self, inst, cls):
        return top.Construct(self.typeName, tuple(arg[inst._idx] for arg in self.args)).result


class altProxy:  # grouped: instance has the branch map
    def __init__(self, name, op):
        self.name = name
        self.op = op

    def __get__(self, inst, cls):
        return inst.brMap.get(self.name, self.op).result


class altItemProxy:  # collection carries branch map (instance comes from collection.__getitem__)
    def __init__(self, name, op):
        self.name = name
        self.op = op

    def __get__(self, inst, cls):
        return inst._parent.brMap.get(self.name, self.op).result[inst._idx]


# internal, leaf/group/collection
def _makeAltClassAndMaps(name, dict_orig, vari, getCol=lambda op: op, attCls=None, altBases=None):
    # vari.getVarName should return the variable and variation name (nomName for the nominal one)
    # if this is a systematic variation branch - otherwise None
    dict_alt = dict(dict_orig)
    # collect ops of kinematic variables that change (nominal as well as varied)
    var_atts = defaultdict(dict)
    for nm, nmAtt in dict_orig.items():
        test = vari.getVarName(nm, collgrpname=name)
        if test is not None:
            attNm, varNm = test
            if normVarName(varNm) in var_atts[attNm]:
                raise RuntimeError(
                    f"Variation {normVarName(varNm)} deduced for {nm} (group {name}), "
                    "but already present")
            var_atts[attNm][normVarName(varNm)] = nmAtt.op
            del dict_alt[nm]
    # redirect in altProxy
    for attNm in var_atts.keys():
        dict_alt[attNm] = attCls(attNm, dict_orig[attNm].op)
    cls_alt = type(f"Alt{name}Proxy", altBases, dict_alt)
    # construct the map of maps of redirections { variation: { varName: op } }
    brMapMap = {}
    for attNm, vAtts in var_atts.items():
        for var, vop in vAtts.items():
            if var not in brMapMap:
                brMapMap[var] = {}
            brMapMap[var][attNm] = getCol(vop)
    # nominal: with systematic variations (all are valid, but not all need to modify)
    nomName = vari.origName if vari.isCalc else vari.nomName(name)
    exclVars = list(normVarName(var) for var in vari.exclVars(name))
    allVars = list(k for k in brMapMap.keys() if k not in exclVars and k != nomName)
    for attNm, vAtts in var_atts.items():
        if nomName not in vAtts:
            raise RuntimeError(f"Nominal variation {nomName} not found for {attNm} (in {vAtts.keys()})")
    brMapMap["nomWithSyst"] = {
        attNm: top.SystAltOp(
            getCol(vAtts[nomName]),
            {var: getCol(vop) for var, vop in vAtts.items() if var not in exclVars and var != nomName},
            valid=tuple(var for var in allVars if var in vAtts),
        ) for attNm, vAtts in var_atts.items()}
    return cls_alt, brMapMap


# Helper classes
class NanoSystematicVarSpec:
    """
    Interface for classes that specify how to incorporate systematics
        or on-the-fly corrections in the decorated tree

    See :py:class:`~.NanoAODDescription` and :py:func:`~.decorateNanoAOD`
    """
    def __init__(self, nomName=None, origName=None, exclVars=None, isCalc=False):
        """ Base class constructor

        :param nomName: nominal variation name (for non-calculated variations;
            can be customised in subclasses by overriding the :py:meth:`~.nomName` method)
        :param origName: original variation name (for calculated variations)
        :param exclVars: variations that are found but should not be used in automatic systematics
            (can be customised through :py:meth:`~.nomName`)
        :param isCalc: boolean indicating whether variations are calculated or read from alternative branches
        """
        self._nomName = nomName
        self.origName = origName
        self._exclVars = exclVars if exclVars is not None else tuple()
        self.isCalc = isCalc

    def appliesTo(self, name):
        """
        Return true if this systematic variation requires action
            for this variable, group, or collection
        """
        return False

    def changesTo(self, name):
        """ Return the new name(s) for a collection or group (assuming appliesTo(name) is True) """
        return (name,)

    def getVarName(self, branchName, collgrpname=None):
        """
        Get the variable name and variation corresponding to an
            (unprefixed, in case of groups or collections) branch name """
        pass

    def nomName(self, name):
        """ Nominal systematic variation name for a group/collection """
        return self._nomName

    def exclVars(self, name):
        """ Systematic variations to exclude for a group/collection """
        return self._exclVars


class ReadVariableVarWithSuffix(NanoSystematicVarSpec):
    """ Read variations of a single branch from branches with the same name with a suffix """
    def __init__(self, commonName, sep="_", nomName="nominal", exclVars=None):
        super().__init__(nomName=nomName, exclVars=exclVars, isCalc=False)
        self.prefix = commonName
        self.sep = sep

    def appliesTo(self, name):
        """ True if name starts with the prefix """
        return name.startswith(self.prefix)

    def getVarName(self, branchName, collgrpname=None):
        """ Split into prefix and variation (if present, else nominal) """
        variNm = normVarName(branchName[len(self.prefix):].lstrip(self.sep))
        return self.prefix, variNm if variNm else self.nomName(collgrpname)


nanoPUWeightVar = ReadVariableVarWithSuffix("puWeight")


class ReadJetMETVar(NanoSystematicVarSpec):
    """
    Read jet and MET kinematic variations from different branches for automatic systematic uncertainties

    :param jetsName: jet collection prefix (e.g. ``"Jet"``)
    :param metName: MET prefix (e.g. ``"MET"``)
    :param jetsNomName: name of the nominal jet variation (``"nom"`` by default)
    :param jetsOrigName: name of the original jet variation (``"raw"`` by default)
    :param metNomName: name of the nominal jet variation (``"nom"`` by default)
    :param metOrigName: name of the original jet variation (``"raw"`` by default)
    :param jetsExclVars: jet variations that are present but should be ignored
        (if not specified, only ``jetsOrigName`` is taken, so if specified
        this should usually be added explicitly)
    :param metExclVars: MET variations that are present but should be ignored
        (if not specified, only ``metOrigName`` is taken, so if specified
        this should usually be added explicitly)
    :param bTaggers: list of b-tagging algorithms, for scale factors stored in a branch
    :param bTagWPs: list of b-tagging working points, for scale factors stored in a branch
        (``shape`` should be included here, if wanted)

    .. note:: The implementation of automatic systematic variations treats
       "xyzup" and "xyzdown" independently (since this is the most flexible).
       If a source of systematic uncertainty should be excluded, both the "up"
       and "down" variation should then be added to the list of variations to
       exclude (``jetsExclVars`` or ``metExclVars``).
    """
    def __init__(
            self, jetsName, metName, jetsNomName="nom", jetsOrigName="raw",
            metNomName="", metOrigName="raw", jetsExclVars=None, metExclVars=None,
            bTaggers=None, bTagWPs=None):
        super().__init__(
            nomName=jetsNomName, origName=jetsOrigName, isCalc=False,
            exclVars=(jetsExclVars if jetsExclVars is not None else (jetsOrigName,)))
        self.jetsName = jetsName
        self.metName = metName
        self.metOrigname = metOrigName
        self.bTaggers = bTaggers if bTaggers is not None else []
        self.bTagWPs = bTagWPs if bTagWPs is not None else []
        self.metNomName = metNomName
        self.metExclVars = (metExclVars if metExclVars is not None else (metOrigName,))

    def appliesTo(self, name):
        return name in (self.jetsName, self.metName)

    def nomName(self, name):
        if name == self.metName:
            return self.metNomName
        else:
            return self._nomName

    def exclVars(self, name):
        if name == self.metName:
            return self.metExclVars
        else:
            return self._exclVars

    def getVarName(self, nm, collgrpname=None):
        if nm.split("_")[0] in ("pt", "eta", "phi", "mass"):
            if len(nm.split("_")) >= 2:
                return (nm.split("_")[0], "_".join(nm.split("_")[1:]))
            # Drop no-suffix (stored) variations if configured to do so
            elif collgrpname == self.jetsName:
                if nm in ("eta", "phi") or self._nomName == "":
                    return nm, self._nomName
                else:  # skip
                    return None
            elif collgrpname == self.metName:
                if self.metNomName == "":
                    return nm, ""
                else:  # skip
                    return None
            else:
                raise RuntimeError(f"Cannot interpret {nm} as a variation of a jet or MET kinematic quantity")
        elif nm.startswith("btagSF"):
            for tagger in self.bTaggers:
                for wp in self.bTagWPs:
                    sfName = f"btagSF_{tagger}_{wp}"
                    if not nm.startswith(sfName):
                        continue
                    if nm == sfName:
                        return nm, "nom"
                    upOrDown = "up" if "up" in nm else "down"
                    if wp != "shape":  # b-tag SF systematics
                        return sfName, f"{sfName}{upOrDown}"
                    else:
                        syst = nm.split(f"{sfName}_{upOrDown}_")[1]
                        if "jes" not in syst:  # b-tag shape systematics
                            return sfName, f"{sfName}_{syst}{upOrDown}"
                        else:  # jes systematics
                            if syst == "jes":
                                syst = "jesTotal"
                            return sfName, f"{syst}{upOrDown}"


nanoReadJetMETVar = ReadJetMETVar(
    "Jet", "MET",
    bTaggers=["csvv2", "deepcsv", "deepjet", "cmva"], bTagWPs=["L", "M", "T", "shape"])
nanoReadJetMETVar_METFixEE2017 = ReadJetMETVar(
    "Jet", "METFixEE2017",
    bTaggers=["csvv2", "deepcsv", "deepjet", "cmva"], bTagWPs=["L", "M", "T", "shape"])


class NanoReadRochesterVar(NanoSystematicVarSpec):
    """
    Read precalculated Rochester correction variations

    :param systName: name of the systematic uncertainty, if variations should be enabled
    """
    def __init__(self, systName=None):
        super().__init__(nomName="nom", origName="orig",
                         exclVars=("orig",), isCalc=False)
        self.systName = systName

    def appliesTo(self, name):
        return name == "Muon"

    def getVarName(self, nm, collgrpname=None):
        if nm.endswith("_pt") and nm.startswith("corrected"):
            var = nm.split("_")[0][len("corrected"):]
            if var == "":
                return "pt", self._nomName
            elif self.systName is not None:
                return "pt", "".join(self.systName, var.lower())
        elif nm == "pt":
            return nm, "orig"


class CalcCollectionsGroups(NanoSystematicVarSpec):
    """
    :py:class:`~.NanoSystematicVarSpec` for on-the-fly corrections
        and systematic variation calculation
    """
    def __init__(self, nomName="nominal", origName="raw", exclVars=None, changes=None, **colsAndAttrs):
        super().__init__(
            nomName=nomName, origName=origName, isCalc=True,
            exclVars=(exclVars if exclVars is not None else (origName,)))
        self.colsAndAttrs = dict()
        for colgrp, attrs in colsAndAttrs.items():
            if isinstance(attrs, str):
                attrs = (attrs,)
            self.colsAndAttrs[colgrp] = attrs
        self.changes = (changes if changes is not None else dict())

    def appliesTo(self, name):
        return name in self.colsAndAttrs

    def changesTo(self, name):
        if name in self.changes:
            return self.changes[name]
        else:
            return super().changesTo(name)

    def getVarName(self, nm, collgrpname=None):
        if collgrpname in self.colsAndAttrs and nm in self.colsAndAttrs[collgrpname]:
            return nm, self.origName


nanoRochesterCalc = CalcCollectionsGroups(Muon="pt")
nanoJetMETCalc = CalcCollectionsGroups(Jet=("pt", "mass"), MET=("pt", "phi"))
nanoJetMETCalc_METFixEE2017 = CalcCollectionsGroups(Jet=("pt", "mass"), METFixEE2017=("pt", "phi"))
nanoFatJetCalc = CalcCollectionsGroups(FatJet=("pt", "mass", "msoftdrop"))


class NanoAODDescription:
    """ Description of the expected NanoAOD structure, and configuration for systematics and corrections

    Essentially, a collection of three containers:
       - :py:attr:`~.collections` a list of collections (by the name of the length leaf)
       - :py:attr:`~.groups` a list of non-collection groups (by prefix, e.g. ``HLT_``)
       - :py:attr:`~.systVariations` a list of :py:class:`~.NanoSystematicVarSpec` instances,
            to configure reading systematics variations from branches, or calculating them on the fly

    The recommended way to obtain a configuration is from the factory method :py:meth:`~.get`
    """
    ProductionVersions = dict()

    def __init__(self, groups=None, collections=None, systVariations=None):
        self.groups = list(groups) if groups is not None else []
        self.collections = list(collections) if collections is not None else []
        self.systVariations = list(systVariations) if systVariations is not None else []

    def clone(self, addGroups=None, removeGroups=None,
              addCollections=None, removeCollections=None, systVariations=None):
        groups = self.groups
        if removeGroups:
            groups = [grp for grp in groups if grp not in removeGroups]
        if addGroups:
            groups = groups + [grp for grp in addGroups if grp not in groups]
        collections = self.collections
        if removeCollections:
            collections = [grp for grp in collections if grp not in removeCollections]
        if addCollections:
            collections = collections + [grp for grp in addCollections if grp not in collections]
        return self.__class__(groups=groups, collections=collections, systVariations=systVariations)

    @staticmethod
    def get(tag, year="2016", isMC=False, addGroups=None, removeGroups=None,
            addCollections=None, removeCollections=None, systVariations=None):
        """ Create a suitable NanoAODDescription instance based on a production version

        A production version is defined by a tag, data-taking year, and a flag
        to distinguish data from simulation.
        Any number of groups or collections can be added or removed from this.
        The ``systVariations`` option

        :Example:

        >>> decorateNanoAOD(tree, NanoAODDescription.get(
        >>>     "v5", year="2016", isMC=True,
        >>>     systVariations=[nanoRochesterCalc, nanoJetMETCalc]))
        >>> decorateNanoAOD(tree, NanoAODDescription.get(
        >>>     "v5", year="2017", isMC=True,
        >>>     systVariations=[nanoPUWeightVar, nanoReadJetMETVar_METFixEE2017]))

        :param tag: production version (e.g. "v5")
        :param year: data-taking year
        :param isMC: simulation or not
        :param addGroups: (optional) list of groups of leaves to add
            (e.g. ``["L1_", "HLT_"]``, if not present)
        :param removeGroups: (optional) list of groups of leaves to remove
            (e.g. ``["L1_"]``, if skimmed)
        :param addCollections: (optional) list of containers to add
            (e.g. ``["nMyJets"]``)
        :param removeCollections: (optional) list of containers to remove
            (e.g. ``["nPhoton", "nTau"]``)
        :param systVariations: list of correction or systematic variation on-the-fly calculators
            or configurations to add (:py:class:`~.NanoSystematicVarSpec` instances)

        See also :py:func:`~.decorateNanoAOD`
        """
        return NanoAODDescription.ProductionVersions[(tag, year, isMC)].clone(
            addGroups=addGroups, removeGroups=removeGroups,
            addCollections=addCollections, removeCollections=removeCollections,
            systVariations=systVariations)


_ndpv = NanoAODDescription.ProductionVersions
_ndpv[("v5", "2016", False)] = NanoAODDescription(
    groups=["CaloMET_", "ChsMET_", "MET_", "PV_", "PuppiMET_", "RawMET_", "TkMET_", "Flag_", "HLT_", "L1_"],
    collections=["nElectron", "nFatJet", "nIsoTrack", "nJet", "nMuon", "nOtherPV", "nPhoton", "nSV",
                 "nSoftActivityJet", "nSubJet", "nTau", "nTrigObj", "nCorrT1METJet"])
_ndpv[("v5", "2016", True)] = _ndpv[("v5", "2016", False)].clone(
    addGroups=["GenMET_", "Generator_", "LHE_", "HTXS_"],
    addCollections=["nGenDressedLepton", "nGenJet", "nGenJetAK8", "nGenPart", "nGenVisTau", "nSubGenJetAK8"])
_ndpv[("v5", "2017", False)] = _ndpv[("v5", "2016", False)].clone(addGroups=["METFixEE2017_"])
_ndpv[("v5", "2017", True)] = _ndpv[("v5", "2016", True)].clone(addGroups=["METFixEE2017_"])
_ndpv[("v5", "2018", False)] = _ndpv[("v5", "2016", False)]
_ndpv[("v5", "2018", True)] = _ndpv[("v5", "2016", True)]
# v6 and v7 have the same structure as v5
for nano_v in ("v6", "v7"):
    for year in ("2016", "2017", "2018"):
        for isMC in (True, False):
            _ndpv[(nano_v, year, isMC)] = _ndpv[("v5", year, isMC)]

nanoGenDescription = NanoAODDescription(
    groups=["Generator_", "MET_", "GenMET_", "LHE_", "HTXS_", "GenVtx_"],
    collections=["nGenPart", "nGenJet", "nGenJetAK8", "nGenVisTau",
                 "nGenDressedLepton", "nGenIsolatedPhoton", "nLHEPart"]
)


def decorateNanoAOD(aTree, description=None):
    """ Decorate a CMS NanoAOD Events tree

    Variation branches following the NanoAODTools conventions (e.g. Jet_pt_nom)
    are automatically used (but calculators for the same collection take
    precendence, if requested).

    :param aTree: TTree to decorate
    :param description: description of the tree format, and configuration for reading
        or calculating systematic variations and corrections,
        a :py:class:`~.NanoAODDescription` instance
        (see also :py:meth:`.NanoAODDescription.get`)
    """
    if description is None:
        raise ValueError(
            'A description is needed to correctly decorate a NanoAOD '
            '(but it may be as simple as ``description=NanoAODDescription.get("v5")``)')

    class GetItemRefCollection:
        def __init__(self, name):
            self.name = name
            self._parent = None

        def __call__(self, me):
            return getattr(self._parent, self.name)

    class GetItemRefCollection_toVar:
        def __init__(self, name):
            self.name = name
            self._parent = None

        def __call__(self, me):
            return getattr(self._parent, self.name).orig

    def addP4ToObj(prefix, lvNms):
        return funProxy(partial((
            lambda getEta, getM, inst: top.Construct(
                "ROOT::Math::LorentzVector<ROOT::Math::PtEtaPhiM4D<float> >",
                (inst.pt, getEta(inst), inst.phi, getM(inst))).result),
            ((lambda inst: inst.eta) if f"{prefix}eta" in lvNms else (lambda inst: 0.)),
            ((lambda inst: inst.mass) if f"{prefix}mass" in lvNms else (lambda inst: 0.))))

    allTreeLeafs = {lv.GetName(): lv for lv in allLeafs(aTree)}
    tree_dict = {"__doc__": f"{aTree.GetName()} tree proxy class"}
    from .root import gbl
    for lvNm, lv in allTreeLeafs.items():
        lvOp = None
        if isinstance(lv, gbl.TBranchElement):  # only RVec (from Snapshot), guess length from name
            if "_" in lvNm:
                guess = f'n{lvNm.split("_")[0]}'
                if guess in allTreeLeafs:
                    tpNm = lv.GetTypeName()
                    m = tpx.vecPat.match(tpNm)
                    if m:  # read as RVec
                        tpNm = f"ROOT::VecOps::RVec<{m.group('item')}>"
                    lvOp = top.GetColumn(tpNm, lvNm)
                else:
                    raise RuntimeError(f"Could not guess the size branch for {lvNm}")
            else:
                raise RuntimeError(f"Not supported")
        else:  # all branches in original NanoAOD
            if not lv.GetLeafCount():  # not an array
                lvOp = top.GetColumn(lv.GetTypeName(), lvNm)
            else:  # count leaf -> array
                cntName = lv.GetLeafCount().GetName()
                lvOp = top.GetArrayColumn(
                    lv.GetTypeName(), lvNm,
                    top.GetColumn(allTreeLeafs[cntName].GetTypeName(), cntName))
        tree_dict[lvNm] = proxy(lvOp)
    tree_children = list()

    def setTreeAtt(name, proxy, setParent=True):
        tree_dict[name] = proxy
        if setParent:
            tree_children.append(proxy)

    # variables with variations
    for vari in description.systVariations:
        toRem = []
        brMap = {}
        attNm = None
        for nm, nmAtt in tree_dict.items():
            if vari.appliesTo(nm):  # only for single-variable variations
                attNm, varNm = vari.getVarName(nm)
                if varNm in brMap:
                    raise RuntimeError(f"Variation {varNm} deduced for {nm}, but already present")
                brMap[varNm] = nmAtt.op
                toRem.append(nm)
        if brMap and (len(brMap) > 1 or vari.isCalc):
            for nm in toRem:
                del tree_dict[nm]
            logger.debug(
                f"Detected systematic variations for variable {attNm} "
                f"(variations: {list(brMap.keys())!r}")
            brMap["nomWithSyst"] = top.SystAltOp(
                brMap[vari.nomName(attNm)], dict(brMap),
                valid=tuple(var for var in brMap.keys() if var != vari.nomName(attNm))
            )
            varsProxy = tpx.AltLeafVariations(None, brMap)
            setTreeAtt(f"_{attNm}", varsProxy)
            setTreeAtt(attNm, varsProxy["nomWithSyst"], False)

    # non-collection branches to group
    grp_found = []
    for prefix in description.groups:
        if not any(lvNm.startswith(prefix) for lvNm in allTreeLeafs):
            logger.warning(f"No branch name starting with {prefix} in the tree - skipping group")
        else:
            grp_found.append(prefix)
    for prefix in grp_found:
        grpNm = prefix.rstrip("_")
        grp_dict = {"__doc__": f"{grpNm} leaf group proxy class"}
        grp_lvNms = {lvNm for lvNm in allTreeLeafs.keys() if lvNm.startswith(prefix)}
        grp_dict.update({
            lvNm[len(prefix):]: proxy(top.GetColumn(allTreeLeafs[lvNm].GetTypeName(), lvNm))
            for lvNm in grp_lvNms})
        if f"{prefix}pt" in grp_lvNms and f"{prefix}phi" in grp_lvNms:
            grp_dict["p4"] = addP4ToObj(prefix, grp_lvNms)
        grpcls = type(f"{grpNm}LeafGroupProxy", (tpx.LeafGroupProxy,), grp_dict)
        for lvNm in grp_lvNms:
            del tree_dict[lvNm]
        # default group proxy, replaced below if needed
        grp_proxy = grpcls(grpNm, None)
        setTreeAtt(grpNm, grp_proxy)
        for vari in description.systVariations:
            if vari.appliesTo(grpNm):
                for newNm in vari.changesTo(grpNm):
                    grpcls_alt, brMapMap = _makeAltClassAndMaps(
                        grpNm, grp_dict, vari, attCls=altProxy,
                        altBases=(tpx.AltLeafGroupProxy,))
                    withSyst = "nomWithSyst"
                    if vari.isCalc:
                        varsProxy = tpx.CalcLeafGroupVariations(
                            None, grp_proxy, brMapMap, grpcls_alt, withSystName=withSyst)
                    else:
                        varsProxy = tpx.AltLeafGroupVariations(None, grp_proxy, brMapMap, grpcls_alt)
                        allVars = list(set(chain.from_iterable(op.variations
                                           for op in varsProxy[withSyst].brMap.values())))
                        if allVars:
                            logger.debug(f"{newNm} variations read from branches: {allVars}")
                    setTreeAtt(f"_{newNm}", varsProxy)
                    setTreeAtt(newNm, varsProxy[withSyst])

    class NanoAODGenRanges:
        @property
        def parent(self):
            return self.genPartMother

        @property
        def ancestors(self):
            return tpx.SelectionProxy(
                self._parent,
                top.Construct(
                    "rdfhelpers::gen::ancestors<-1,ROOT::VecOps::RVec<Int_t>>",
                    (tf.static_cast("int", self.genPartMother._idx), self.genPartMother._idx.arg)),
                type(self))

    # SOA, nanoAOD style (LeafCount, shared)
    cnt_found = []
    for sizeNm in description.collections:
        if sizeNm not in allTreeLeafs:
            logger.warning(f"{sizeNm} is not a branch in the tree - skipping collection")
        else:
            cnt_found.append(sizeNm)

    for sizeNm in cnt_found:
        grpNm = sizeNm[1:]
        prefix = f"{grpNm}_"
        itm_dict = {"__doc__": f"{grpNm} proxy class"}
        itm_lvs = {lvNm for lvNm, lv in allTreeLeafs.items()
                   if lvNm.startswith(prefix) and (isinstance(lv, gbl.TBranchElement)
                                                   or lv.GetLeafCount().GetName() == sizeNm)}
        sizeOp = top.GetColumn(allTreeLeafs[sizeNm].GetTypeName(), sizeNm)
        for lvNm in itm_lvs:
            lvNm_short = lvNm[len(prefix):]
            if isinstance(allTreeLeafs[lvNm], gbl.TBranchElement):
                col = tree_dict[lvNm].op.result
            else:
                col = top.GetArrayColumn(allTreeLeafs[lvNm].GetTypeName(), lvNm, sizeOp).result
            if "Idx" not in lvNm:
                itm_dict[lvNm_short] = itemProxy(col)
            else:
                coll, i = lvNm_short.split("Idx")
                collPrefix = coll[0].capitalize() + coll[1:]
                collGetter = (
                    GetItemRefCollection_toVar(f"_{collPrefix}")
                    if any(vari.appliesTo(collPrefix) for vari in description.systVariations)
                    else GetItemRefCollection(collPrefix))
                tree_children.append(collGetter)
                itm_dict["".join((coll, i))] = itemRefProxy(col, collGetter)
        # create p4 branches (naive, but will be reused for variation case)
        if f"{prefix}pt" in itm_lvs and f"{prefix}phi" in itm_lvs:
            itm_dict["p4"] = addP4ToObj(prefix, itm_lvs)
        itm_bases = [tpx.ContainerGroupItemProxy]
        if sizeNm == "nGenPart":
            itm_bases.append(NanoAODGenRanges)
        itmcls = type(f"{grpNm}GroupItemProxy", tuple(itm_bases), itm_dict)
        # default collection proxy, replaced below if needed
        coll_orig = tpx.ContainerGroupProxy(prefix, None, sizeOp, itmcls)
        setTreeAtt(grpNm, coll_orig)
        # insert variations using kinematic calculator, from branches, or not
        for vari in description.systVariations:
            if vari.appliesTo(grpNm):
                altItemType, brMapMap = _makeAltClassAndMaps(
                    grpNm, itm_dict, vari, getCol=(lambda att: att.op),
                    attCls=altItemProxy, altBases=(tpx.ContainerGroupItemProxy,))
                withSyst = "nomWithSyst"
                for newNm in vari.changesTo(grpNm):
                    if vari.isCalc:
                        varsProxy = tpx.CalcCollectionVariations(
                            None, coll_orig, brMapMap, altItemType=altItemType, withSystName=withSyst)
                    else:
                        varsProxy = tpx.AltCollectionVariations(
                            None, coll_orig, brMapMap, altItemType=altItemType)
                        allVars = list(set(chain.from_iterable(
                            op.variations for op in varsProxy[withSyst].brMap.values())))
                        if allVars:
                            logger.debug(f"{newNm} variations read from branches: {allVars}")
                    setTreeAtt(f"_{newNm}", varsProxy)
                    setTreeAtt(newNm, varsProxy[withSyst])

        for lvNm in itm_lvs:
            del tree_dict[lvNm]
        del tree_dict[sizeNm]  # go through tf.rng_len

    TreeProxy = type(f"{aTree.GetName()}Proxy", (tpx.TreeBaseProxy,), tree_dict)
    treeProxy = TreeProxy(aTree)
    for pc in tree_children:
        pc._parent = treeProxy

    return treeProxy


def decorateCMSPhase2SimTree(aTree, isMC=True):
    """ Decorate a flat tree as used for CMS Phase2 physics studies """

    class Phase2GenRanges:
        @property
        def parent(self):
            return self._parent[self.m1]

        @property
        def ancestors(self):
            return tpx.SelectionProxy(
                self._parent,
                top.Construct(
                    "rdfhelpers::gen::ancestors<-99>",
                    (tf.static_cast("int", self.parent._idx), self.parent._idx.arg)),
                type(self))

        @property
        def children(self):
            return self._parent[self.d1:tf.switch(self.d2 > 0, self.d2 + 1, self.d1)]

        @property
        def descendants(self):
            return tpx.SelectionProxy(
                self._parent,
                top.Construct(
                    "rdfhelpers::gen::descendants_firstlast<-99,ROOT::VecOps::RVec<Int_t>>",
                    (tf.static_cast("int", self._idx), self.d1._parent.arg, self.d2._parent.arg)),
                type(self))

    allTreeLeafs = {lv.GetName(): lv for lv in allLeafs(aTree)}
    tree_dict = {"__doc__": f"{aTree.GetName()} tree proxy class"}
    # fill all, take some out later
    tree_dict.update({lvNm: proxy(top.GetColumn(lv.GetTypeName(), lvNm))
                      for lvNm, lv in allTreeLeafs.items()})
    tree_children = []
    cnt_lvs = {lv.GetLeafCount().GetName()
               for lv in allTreeLeafs.values() if lv.GetLeafCount()}
    for sizeNm in cnt_lvs:
        grpNm = sizeNm.split("_")[0]
        prefix = f"{grpNm}_"
        itm_dict = {"__doc__": f"{grpNm} proxy class"}
        itm_lvs = {lvNm for lvNm, lv in allTreeLeafs.items()
                   if lvNm.startswith(prefix) and lvNm != sizeNm
                   and lv.GetLeafCount().GetName() == sizeNm}
        sizeOp = top.GetColumn(allTreeLeafs[sizeNm].GetTypeName(), sizeNm)
        if allTreeLeafs[sizeNm].GetTypeName() != top.SizeType:
            sizeOp = top.adaptArg(tf.static_cast(top.SizeType, sizeOp))
        for lvNm in itm_lvs:
            lvNm_short = lvNm[len(prefix):]
            col = top.GetArrayColumn(allTreeLeafs[lvNm].GetTypeName(), lvNm, sizeOp).result
            itm_dict[lvNm_short] = itemProxy(col)
        # create p4 branches
        p4AttNames = ("pt", "eta", "phi", "mass")
        if all(("".join((prefix, att)) in itm_lvs) for att in p4AttNames):
            itm_dict["p4"] = funProxy(lambda inst: top.Construct(
                "ROOT::Math::LorentzVector<ROOT::Math::PtEtaPhiM4D<float> >",
                (inst.pt, inst.eta, inst.phi, inst.mass)).result)
            # mother/daughter references
        itm_bases = [tpx.ContainerGroupItemProxy]
        if grpNm == "genpart":
            itm_bases.append(Phase2GenRanges)
        itmcls = type(f"{grpNm.capitalize()}GroupItemProxy", tuple(itm_bases), itm_dict)
        coll = tpx.ContainerGroupProxy(prefix, None, sizeOp, itmcls)
        tree_dict[grpNm] = coll
        tree_children.append(coll)

        for lvNm in itm_lvs:
            del tree_dict[lvNm]
        del tree_dict[sizeNm]  # go through tf.rng_len

    TreeProxy = type(f"{aTree.GetName().capitalize()}Proxy", (tpx.TreeBaseProxy,), tree_dict)
    treeProxy = TreeProxy(aTree)

    for pc in tree_children:
        pc._parent = treeProxy

    return treeProxy



import os.path

import pytest

testData = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def decoNano():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016postproc_JMEKin_bTagShape_puWeight.root"))
    tree = f.Get("Events")
    from bamboo.treedecorators import decorateNanoAOD, NanoAODDescription, nanoPUWeightVar, ReadJetMETVar
    from bamboo.dataframebackend import DataframeBackend
    nanoReadJetMETVar_MC = ReadJetMETVar(
        "Jet", "MET_T1",
        bTaggers=["csvv2", "deepcsv", "deepjet", "cmva"],
        bTagWPs=["L", "M", "T", "shape"])
    tup = decorateNanoAOD(
        tree, NanoAODDescription.get(
            "v5", year="2016", isMC=True,
            systVariations=[nanoPUWeightVar, nanoReadJetMETVar_MC]))
    be, noSel = DataframeBackend.create(tup)
    yield tup


def test_getSimpleObjects(decoNano):
    assert decoNano.Electron[0]
    assert decoNano.Muon[0]
    assert decoNano.Photon[0]
    assert decoNano.SubJet[0]
    assert decoNano.FatJet[0]


def test_getJet(decoNano):
    assert decoNano.Jet[0]


def test_simpleRef(decoNano):
    assert decoNano.Electron[0].photon
    assert decoNano.FatJet[0].subJet1


def test_toJetRef(decoNano):
    assert decoNano.Electron[0].jet
    assert decoNano.Muon[0].jet


def test_fromJetRef(decoNano):
    jet = decoNano.Jet[0]
    assert jet.muon1
    assert jet.muon2
    assert jet.electron1
    assert jet.electron2



"""
Example analysis module: electron Tag&Probe from a NanoAOD
"""
import logging

from bamboo.analysismodules import NanoAODHistoModule

logger = logging.getLogger(__name__)


class TagAndProbe1(NanoAODHistoModule):
    def __init__(self, args):
        super().__init__(args)

    def make1D(self, name, variable, selection, binning, **kwargs):
        from bamboo.plots import Plot
        nm = f"{name}_{selection.name}"
        return Plot.make1D(nm, variable, selection, binning, **kwargs)

    def definePlots(self, t, noSel, era=None, sample=None, sampleCfg=None):
        from bamboo.plots import Plot
        from bamboo.plots import EquidistantBinning as EqB
        from bamboo import treefunctions as op
        from itertools import count

        plots = []

        tnpTrigSel = noSel.refine("hasTnPTrig", cut=t.HLT.Ele32_eta2p1_WPTight_Gsf)

        tagEle = op.select(
            t.Electron,
            lambda el: op.AND(
                el.pt >= 30.,
                op.abs(el.eta + el.deltaEtaSC) < 2.1,
                op.NOT(op.in_range(1.4442, op.abs(el.eta + el.deltaEtaSC), 1.566)),
                el.cutBased_Fall17_V1,
                op.rng_any(t.TrigObj, lambda to: op.AND(
                    to.id == 11, to.filterBits & 2,  # TODO review 'filterBits'
                    op.sqrt((to.eta - el.eta)**2 + (to.phi - el.phi)**2) < .4
                ))
            ))

        probeEle = op.select(t.Electron, lambda el: op.abs(el.eta + el.deltaEtaSC) < 2.5)

        plots.append(Plot.make1D(
            "nTagEle", op.rng_len(tagEle), tnpTrigSel, EqB(10, 0., 10.),
            title="Number of tag electrons (partial selection)"))
        plots.append(Plot.make1D(
            "nProbeEle", op.rng_len(probeEle), tnpTrigSel, EqB(10, 0., 10.),
            title="Number of probe electrons (partial selection)"))

        eleTagAndProbe = op.combine(
            (tagEle, probeEle),
            pred=lambda elt, elp: op.in_range(50., op.invariant_mass(elt.p4, elp.p4), 130.),
            samePred=lambda o1, o2: o1.idx != o2.idx)

        plots.append(Plot.make1D(
            "nTnPElePairs", op.rng_len(eleTagAndProbe), tnpTrigSel, EqB(10, 0., 10.),
            title="Tag&Probe electron: number of pairs"))

        hasTnPpair = tnpTrigSel.refine("hasTnPpair", cut=(op.rng_len(eleTagAndProbe) > 0))
        theTnPpair = eleTagAndProbe[0]  # TODO add more strategies
        # theTnpPair = op.rng_pickRandom(tnpPairs, seed=...)
        tag, probe = theTnPpair[0], theTnPpair[1]  # noqa: F841
        tnpMll = op.invariant_mass(theTnPpair[0].p4, theTnPpair[1].p4)
        plots.append(Plot.make1D(
            "eleTnP_ptTag", theTnPpair[0].p4.Pt(), hasTnPpair, EqB(50, 0., 150.),
            title="Tag&Probe electron pair: probe PT"))
        plots.append(Plot.make1D(
            "eleTnP_ptProbe", theTnPpair[1].p4.Pt(), hasTnPpair, EqB(50, 0., 150.),
            title="Tag&Probe electron pair: probe PT"))

        plots.append(Plot.make1D(
            "eleTnPmassInteg", tnpMll, hasTnPpair, EqB(80, 50., 130.),
            title="Tag&Probe electron pair invariant mass"))

        ptBinning = [15., 20., 25., 30., 40., 50., 70., 100., 200.]
        denomSels = {f"PT{i}": hasTnPpair.refine(f"eleTnP_denom_PT{i}",
                                                 cut=op.in_range(ptMn, probe.p4.Pt(), ptMx))
                     for i, ptMn, ptMx in zip(count(), ptBinning[:-1], ptBinning[1:])}
        for dSel in denomSels.values():
            plots.append(Plot.make1D(
                f"{dSel.name}_mll", tnpMll, dSel, EqB(80, 50., 130.),
                title="Tag&Probe electron pair invariant mass"))

        selsToTest = {
            # reading vector<bool> branches needs a fix from ROOT 6.16;
            # 'op.construct' forces copy to bool, this workaround should not be needed
            # (if the bool comes from an expression it isn't)
            # "MVAIso90" : op.construct("bool", (probe.mvaFall17V1Iso_WP90,))
            "mvaTTH090": probe.mvaTTH > 0.95  # for demonstration only
        }
        for tsName, testSel in selsToTest.items():
            for bNm, dSel in denomSels.items():
                nomSel = dSel.refine(f"eleTnP_pass{tsName}_{bNm}", cut=testSel)
                plots.append(Plot.make1D(
                    f"{nomSel.name}_mll", tnpMll, nomSel, EqB(80, 50., 130.),
                    title="Tag&Probe electron pair invariant mass"))

        return plots



"""
Implementation of the IRIS-HEP ADL benchmarks https://github.com/iris-hep/adl-benchmarks-index
using the bamboo framework https://gitlab.cern.ch/cp3-cms/bamboo

After installing bamboo (see the documentation, plotIt is not required) this can be run with
$ bambooRun -m bamboo/examples/adl_benchmarks.py:IRISHEP_ADLBenchmarks
    bamboo/examples/adl_benchmarks.yml -o adl_bamboo
the resulting histograms will be stored in `adl_bamboo/results/SingleMuon_test.root`.
The `--examples` argument can be used to run just some of the benchmarks (default is all);
since some of the benchmarks go through a lot of combinatorics, implicit multithreading
is useful to speed them up, and can be enabled by passing `--threads N` (with N=5 or so).
"""
__author__ = "Pieter David <pieter.david@cern.ch>"

import logging

from bamboo.analysismodules import NanoAODHistoModule

logger = logging.getLogger(__name__)


class IRISHEP_ADLBenchmarks(NanoAODHistoModule):
    def postProcess(self, taskList, config=None, workdir=None, resultsdir=None):
        # disabled, no need for plotIt then
        logger.info(f"The resulting histograms are stored in the {resultsdir} directory")

    def mergeCounters(self, outF, infileNames, sample=None):
        pass  # disabled because the test file is not a full NanoAOD

    def addArgs(self, parser):
        super().addArgs(parser)
        self.nExamples = 8
        parser.add_argument(
            "--examples", type=str, default="all",
            help=(f'Set of examples to run (comma-separated, any set of 1-{self.nExamples:d}), '
                  'or "all" (default)'))

    def definePlots(self, tree, noSel, sample=None, sampleCfg=None):
        from bamboo.plots import Plot, SummedPlot
        from bamboo.plots import EquidistantBinning as EqBin
        from bamboo import treefunctions as op

        if self.args.examples == "all":
            examples = list(i + 1 for i in range(self.nExamples))  # 1-4 are fine, so is 7
        else:
            examples = list({int(tk) for tk in self.args.examples.split(",")})
        logger.info("Running the following examples: {}".format(",".join(str(i) for i in examples)))

        plots = []
        if 1 in examples:
            # Example 1: Plot the missing ET of all events.
            plots.append(Plot.make1D(
                "Ex1_MET", tree.MET.sumet, noSel,
                EqBin(100, 0., 2000.), title="MET (GeV)"))

        if 2 in examples:
            # Example 2: Plot pT of all jets in all events.
            plots.append(Plot.make1D(
                "Ex2_jetPt", op.map(tree.Jet, lambda j: j.pt), noSel,
                EqBin(100, 15., 60.), title="Jet p_{T} (GeV/c)"))

        if 3 in examples:
            # Example 3: Plot pT of jets with |η| < 1.
            centralJets1 = op.select(tree.Jet, lambda j: op.abs(j.eta) < 1.)
            plots.append(Plot.make1D(
                "Ex3_central1_jetPt", op.map(centralJets1, lambda j: j.pt), noSel,
                EqBin(100, 15., 60.), title="Jet p_{T} (GeV/c)"))

        if 4 in examples:
            # Example 4: Plot the missing ET of events that have at least two jets with pT > 40 GeV.
            jets40 = op.select(tree.Jet, lambda j: j.pt > 40)
            hasTwoJets40 = noSel.refine("twoJets40", cut=(op.rng_len(jets40) >= 2))
            plots.append(Plot.make1D(
                "Ex4_twoJets40_MET", tree.MET.sumet, hasTwoJets40,
                EqBin(100, 0., 2000.), title="MET (GeV)"))

        if 5 in examples:
            # Example 5: Plot the missing ET of events that have an opposite-sign muon pair
            # with an invariant mass between 60 and 120 GeV.
            dimu_Z = op.combine(
                tree.Muon, N=2, pred=(lambda mu1, mu2: op.AND(
                    mu1.charge != mu2.charge,
                    op.in_range(60., op.invariant_mass(mu1.p4, mu2.p4), 120.)
                )))
            hasDiMuZ = noSel.refine("hasDiMuZ", cut=(op.rng_len(dimu_Z) > 0))
            plots.append(Plot.make1D(
                "Ex5_dimuZ_MET", tree.MET.sumet, hasDiMuZ, EqBin(100, 0., 2000.), title="MET (GeV)"))

        if 6 in examples:
            # Example 6: Plot pT of the trijet system with the mass closest to 172.5 GeV in each event
            # and plot the maximum b-tagging discriminant value among the jets in the triplet.
            plots.append(Plot.make1D(
                "Ex6_njets", op.rng_len(tree.Jet), noSel,
                EqBin(20, 0., 20.), title="Number of jets"))
            trijets = op.combine(tree.Jet, N=3)
            plots.append(Plot.make1D(
                "Ex6_ntrijets", op.rng_len(trijets), noSel,
                EqBin(100, 0., 1000.), title="Number of 3-jet combinations"))
            hadTop = op.rng_min_element_by(
                trijets, fun=lambda comb: op.abs((comb[0].p4 + comb[1].p4 + comb[2].p4).M() - 172.5))
            hasTriJet = noSel.refine("hasTriJet", cut=(op.rng_len(trijets) > 0))
            plots.append(Plot.make1D(
                "Ex6_trijet_mass", (hadTop[0].p4 + hadTop[1].p4 + hadTop[2].p4).M(), hasTriJet,
                EqBin(100, 0., 250.), title="Trijet mass (GeV/c^{2})"))
            plots.append(Plot.make1D(
                "Ex6_trijet_topPt", (hadTop[0].p4 + hadTop[1].p4 + hadTop[2].p4).Pt(), hasTriJet,
                EqBin(100, 0., 250.), title="Trijet p_{T} (GeV/c)"))
            plots.append(Plot.make1D(
                "Ex6_trijet_maxbtag",
                op.max(op.max(hadTop[0].btag, hadTop[1].btag), hadTop[2].btag), hasTriJet,
                EqBin(100, 0., 1.), title="Trijet maximum b-tag"))

        if 7 in examples:
            # Example 7: Plot the sum of pT of jets with pT > 30 GeV
            # that are not within 0.4 in ΔR of any lepton with pT > 10 GeV.
            el10 = op.select(tree.Electron, lambda el: el.pt > 10.)
            mu10 = op.select(tree.Muon, lambda mu: mu.pt > 10.)
            cleanedJets30 = op.select(
                tree.Jet, lambda j: op.AND(
                    j.pt > 30.,
                    op.NOT(op.rng_any(el10, lambda el: op.deltaR(j.p4, el.p4) < 0.4)),
                    op.NOT(op.rng_any(mu10, lambda mu: op.deltaR(j.p4, mu.p4) < 0.4))
                ))
            plots.append(Plot.make1D(
                "Ex7_sumCleanedJetPt", op.rng_sum(cleanedJets30, lambda j: j.pt), noSel,
                EqBin(100, 15., 200.), title="Sum p_{T} (GeV/c)"))

        if 8 in examples:
            # Example 8: For events with at least three leptons and a same-flavor opposite-sign lepton pair,
            # find the same-flavor opposite-sign lepton pair with the mass closest to 91.2 GeV
            # and plot the transverse mass of the missing energy and the leading other lepton.
            # The plot is made for each of the different flavour categories (l+/- l-/+ l') and then summed,
            # because concatenation of containers is not (yet) supported.
            lepColl = {"El": tree.Electron, "Mu": tree.Muon}
            mt3lPlots = []
            for dlNm, dlCol in lepColl.items():
                dilep = op.combine(dlCol, N=2, pred=(lambda l1, l2: op.AND(l1.charge != l2.charge)))
                hasDiLep = noSel.refine("hasDilep{0}{0}".format(dlNm), cut=(op.rng_len(dilep) > 0))
                dilepZ = op.rng_min_element_by(
                    dilep, fun=lambda ll: op.abs(op.invariant_mass(ll[0].p4, ll[1].p4) - 91.2))
                for tlNm, tlCol in lepColl.items():
                    if tlCol == dlCol:
                        hasTriLep = hasDiLep.refine("hasTrilep{0}{0}{1}".format(dlNm, tlNm),
                                                    cut=(op.rng_len(tlCol) > 2))
                        residLep = op.select(
                            tlCol, lambda l: op.AND(l.idx != dilepZ[0].idx, l.idx != dilepZ[1].idx))
                        l3 = op.rng_max_element_by(residLep, lambda l: l.pt)
                    else:
                        hasTriLep = hasDiLep.refine("hasTriLep{0}{0}{1}".format(dlNm, tlNm),
                                                    cut=(op.rng_len(tlCol) > 0))
                        l3 = op.rng_max_element_by(tlCol, lambda l: l.pt)
                    mtPlot = Plot.make1D(
                        "Ex8_3lMT_{0}{0}{1}".format(dlNm, tlNm),
                        op.sqrt(2 * l3.pt * tree.MET.pt * (1 - op.cos(l3.phi - tree.MET.phi))),
                        hasTriLep, EqBin(100, 0., 250.), title="M_{T} (GeV/c^2)")
                    mt3lPlots.append(mtPlot)
                    plots.append(mtPlot)
            plots.append(SummedPlot("Ex8_3lMT", mt3lPlots))

        return plots



""" Setuptools-based setup module for bamboo """
import os
import os.path

import pkg_resources
from setuptools import find_packages

from skbuild import setup

try:
    from sphinx.setup_command import BuildDoc
except ImportError:
    BuildDoc = None

cmake_args = []
if "VIRTUAL_ENV"in os.environ:  # allow picking up packages from virtualenv
    cmake_args.append("-DCMAKE_PREFIX_PATH={}".format(os.environ["VIRTUAL_ENV"]))
try:  # pick up pytorch from environment (needs --no-build-isolation)
    torch_cmake_locPath = os.path.join("share", "cmake", "Torch")
    if (pkg_resources.resource_exists("torch", torch_cmake_locPath)
        and (pkg_resources.get_distribution("torch").parsed_version
             >= pkg_resources.parse_version("1.2.0"))):
        cmake_args.append("-DTorch_DIR={}".format(
            pkg_resources.resource_filename("torch", torch_cmake_locPath)))
except ImportError:
    print(
        "Warning: optional dependency torch (>=1.2.0) not found. "
        "Install bamboo[torch] to install it as a dependency, "
        "or use --no-build-isolation to pick up an existing install from the environment")

setup(
    packages=find_packages(".", exclude=["ext", "tests", "examples"]),
    cmake_source_dir="ext",
    cmake_args=cmake_args,
    cmdclass={
        "build_sphinx": BuildDoc
    }
)



"""
The :py:mod:`bamboo.root` module collects a set of thin wrappers around ROOT
methods, and centralizes the import of the Cling interpreter global namespace
in PyROOT. For compatibility, it is recommended that user code uses
``from bamboo.root import gbl`` rather than ``import ROOT as gbl`` or
``from cppyy import gbl``.
"""
import functools
from contextlib import contextmanager

import pkg_resources

import ROOT; gbl = ROOT  # noqa: E702
_rootVersion = gbl.gROOT.GetVersion()
_rootVersion_split = tuple(int(iv) for tk in _rootVersion.split("/") for iv in tk.split("."))
if _rootVersion_split[0] == 6 and _rootVersion_split[1] < 22:
    gbl.PyConfig.IgnoreCommandLineOptions = True


def addIncludePath(incPath):
    """ Add an include path to the ROOT interpreter """
    gbl.gInterpreter.AddIncludePath(incPath)


def loadHeader(headerName):
    """ Include a C++ header in the ROOT interpreter """
    gbl.gROOT.ProcessLine(f'#include "{headerName}"')


def addDynamicPath(libPath):
    """ Add a dynamic library path to the ROOT interpreter"""
    gbl.gSystem.AddDynamicPath(libPath)


def loadLibrary(libName):
    """ Load a shared library in the ROOT interpreter """
    st = gbl.gSystem.Load(libName)
    if st == -1:
        raise RuntimeError(f"Library {libName} could not be found")
    elif st == -2:
        raise RuntimeError(f"Version match for library {libName}")


def findLibrary(libName):
    """ Check if a library can be found, and returns the path in that case """
    ret = gbl.gSystem.FindDynamicLibrary(gbl.TString(libName), True)
    if ret != "":
        return str(ret)


def loadDependency(bambooLib=None, includePath=None, headers=None, dynamicPath=None, libraries=None):
    """
    Load a C++ extension

    :param bambooLib: name(s) of the bamboo extension libraries, if any
    :param includePath: include directory for headers
    :param headers: headers to load explicitly (which can depend on other headers in the inclue path)
    :param dynamicPath: dynamic library path to add
    :param libraries: additional shared libraries to load
    """
    def asList(arg):
        if isinstance(arg, str):
            return [arg]
        else:
            return arg
    if dynamicPath:
        for pth in asList(dynamicPath):
            addDynamicPath(pth)
    if libraries:
        for lib in asList(libraries):
            loadLibrary(f"lib{lib}")
    if includePath:
        for pth in asList(includePath):
            addIncludePath(pth)
    if headers:
        for hdr in asList(headers):
            loadHeader(hdr)
    if bambooLib:
        for lib in asList(bambooLib):
            loadLibrary(f"lib{lib}")


@contextmanager
def returnToDirectory():
    gpwd = gbl.gDirectory.GetPath()
    yield
    gbl.gDirectory.cd(gpwd)


class once:
    """ Function decorator to make sure things are not loaded more than once """
    def __init__(self, fun):
        self.fun = fun
        self.has_run = False
        functools.update_wrapper(self, fun)

    def __call__(self, **kwargs):
        if not self.has_run:
            self.has_run = True
            return self.fun(**kwargs)


_defaultHeaders = ("Math/VectorUtil.h", "bamboohelpers.h", "range.h",
                   "LumiMask.h", "bamboorandom.h",
                   "scalefactors.h", "BTagCalibrationStandalone.h")
_defaultBambooLibs = ("BambooLumiMask", "BambooRandom", "BinnedValues")
_defaultDependencies = ("ROOT::ROOTDataFrame", "ROOT::GenVector")


@once
def loadBambooExtensions():
    # Add extension libraries and necessary header files to the ROOT interpreter
    incDir = pkg_resources.resource_filename("bamboo", "include")
    libDir = pkg_resources.resource_filename("bamboo", "lib")
    import os.path
    if not (os.path.exists(incDir) and os.path.exists(libDir)):
        raise RuntimeError(
            f"Could not find {incDir} and/or {libDir}, this may happen when running "
            "from the source repository and using a virtualenv with non-editable install")
    addIncludePath(incDir)
    addDynamicPath(libDir)
    # now load default headers and libraries
    for lib in _defaultBambooLibs:
        loadLibrary(f"lib{lib}")
    for fname in _defaultHeaders:
        loadHeader(fname)


@once
def loadJMESystematicsCalculators(backend=None):
    loadBambooExtensions()
    try:
        import CMSJMECalculators
        CMSJMECalculators.loadJMESystematicsCalculators()
    except ImportError:
        raise RuntimeError(
            "Could not load jet&MET correction and systematic variations calculators. "
            "Please install them (see https://gitlab.cern.ch/cp3-cms/CMSJMECalculators "
            "for more information)"
        ) from None
    if backend:
        import pkg_resources
        calc_dir = pkg_resources.resource_filename("CMSJMECalculators", "cmake")
        backend.addDependency(
            package="CMSJMECalculators", cmakeArgs=f"-DCMSJMECalculators_DIR={calc_dir}",
            headers="JMESystematicsCalculators.h", libraries="CMSJMECalculators"
        )


@once
def loadRochesterCorrectionCalculator(backend=None):
    loadBambooExtensions()
    loader = backend.addDependency if backend else loadDependency
    loader(bambooLib="RoccoR", headers="RochesterCorrectionCalculator.h")
    getattr(gbl, "RochesterCorrectionCalculator::result_t")  # trigger dictionary generation


@once
def loadLibTorch(backend=None):
    if loadTensorflowC.has_run:
        raise RuntimeError("If loading both Tensorflow-C and libtorch, the latter should be loaded first")
    loadBambooExtensions()
    torchLib = findLibrary("libtorch")
    dynPath = None
    if (not torchLib) and pkg_resources.resource_filename("torch", "lib"):
        dynPath = pkg_resources.resource_filename("torch", "lib")
    loader = backend.addDependency if backend else loadDependency
    loader(bambooLib="BambooTorch", headers="bambootorch.h", dynamicPath=dynPath)


def _addVirtualEnvPaths():
    import os
    incPath, dynPath = None, None
    if "VIRTUAL_ENV" in os.environ:
        incPath = addIncludePath(os.path.join(os.environ["VIRTUAL_ENV"], "include"))
        dynPath = addDynamicPath(os.path.join(os.environ["VIRTUAL_ENV"], "lib"))
    return incPath, dynPath


@once
def loadTensorflowC(backend=None):
    loadBambooExtensions()
    tfLib = findLibrary("libtensorflow")
    incPath, dynPath = None, None
    if not tfLib:
        incPath, dynPath = _addVirtualEnvPaths()
    loader = backend.addDependency if backend else loadDependency
    loader(bambooLib="BambooTensorflowC", headers="bambootensorflowc.h",
           includePath=incPath, dynamicPath=dynPath, libraries="tensorflow")


@once
def loadlwtnn(backend=None):
    loadBambooExtensions()
    lwtnnLib = findLibrary("liblwtnn")
    incPath, dynPath = None, None
    if not lwtnnLib:
        incPath, dynPath = _addVirtualEnvPaths()
    loader = backend.addDependency if backend else loadDependency
    loader(bambooLib="BambooLwtnn", headers="bamboolwtnn.h",
           includePath=incPath, dynamicPath=dynPath, libraries="lwtnn")


@once
def loadONNXRuntime(backend=None):
    loadBambooExtensions()
    onnxRuntimeLib = findLibrary("libonnxruntime")
    incPath, dynPath = None, None
    if not onnxRuntimeLib:
        incPath, dynPath = _addVirtualEnvPaths()
    onnxRuntimeLib = findLibrary("libonnxruntime")
    import os.path
    incDir = os.path.join(os.path.dirname(os.path.dirname(onnxRuntimeLib)), "include")
    if os.path.isdir(os.path.join(incDir, "onnxruntime")):
        incDir = os.path.join(incDir, "onnxruntime")
    onnx_cxx_hdr = os.path.join(incDir, "core", "session", "onnxruntime_cxx_api.h")
    if not os.path.isfile(onnx_cxx_hdr):
        raise RuntimeError(f"Could not find onnxruntime header {onnx_cxx_hdr}")
    loader = backend.addDependency if backend else loadDependency
    loader(bambooLib="BambooONNXRuntime", headers="bambooonnxruntime.h",
           includePath=os.path.dirname(onnx_cxx_hdr),
           dynamicPath=dynPath, libraries="onnxruntime")


@once
def loadcorrectionlib(backend=None):
    import os.path
    loader = backend.addDependency if backend else loadDependency
    loader(headers="correction.h",
           includePath=pkg_resources.resource_filename("correctionlib", "include"),
           libraries=pkg_resources.resource_filename(
               "correctionlib", os.path.join("lib", "libcorrectionlib.so"))
           )
